# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only
# (subject to the limitations in the disclaimer below)
# provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Standard Library Imports
import tkinter as tk
from pathlib import Path

# Third Party Imports
from PIL import Image, ImageTk

# Local Imports


class SplashScreen(tk.Toplevel):
    """Display Splash Screen

    Briefly shows a splash screen upon loading the software.
    Centered depending upon the host computer being used.
    """

    def __init__(self, root, image_path, *args, **kargs):
        """Initialize the SplashScreen.

        Parameters
        ----------
        root : Tk top-level widget
            Tkinter GUI instance to which this SplashScreen belongs.
        image_path : str
            Path to the image to display on the splash screen.
        *args
            Additional arguments to pass to the tk.Toplevel constructor.
        **kargs
            Additional keyword arguments to pass to the tk.Toplevel constructor.
        """

        tk.Toplevel.__init__(self, root)
        # without navigation panel
        self.overrideredirect(True)

        try:
            img = tk.PhotoImage(file=image_path)
            w, h = img.width(), img.height()  # width, height of the image
            loading_label = tk.Label(self, image=img)
        except tk.TclError:
            w, h = 300, 100
            loading_label = tk.Label(self, text="Loading Navigate Software ...")
        loading_label.pack()

        # get screen width and height
        ws = root.winfo_screenwidth()  # width of the screen
        hs = root.winfo_screenheight()  # height of the screen

        # calculate x and y coordinates for the Tk root window
        x = (ws / 2) - (w / 2)
        y = (hs / 2) - (h / 2)

        # draw the window in the center of the window
        self.geometry("%dx%d+%d+%d" % (w, h, x, y))
        self.resizable(0, 0)
        self.update()




# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only (subject to the
# limitations in the disclaimer below) provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.


# Standard Library Imports
from tkinter import filedialog, messagebox
import math
import logging

# Third Party Imports
import pandas as pd
from pandastable import TableModel

# Local Imports
from navigate.controller.sub_controllers.gui_controller import GUIController


# Logger Setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)


class MultiPositionController(GUIController):
    """Controller for the Multi-Position Acquisition Interface."""

    def __init__(self, view, parent_controller=None):
        """Initialize the Multi-Position Acquisition Interface.

        Parameters
        ----------
        view : MultiPositionView
            view for the Multi-Position Acquisition Interface
        parent_controller : Controller, optional
            parent controller, by default None
        """
        super().__init__(view, parent_controller)
        #: MultiPositionTable: Multi-Position Acquisition Interface
        self.table = self.view.pt
        # self.table.rowheader.bind("<Double-Button-1>", self.handle_double_click)
        self.table.loadCSV = self.load_positions
        self.table.exportCSV = self.export_positions
        self.table.insertRow = self.insert_row_func
        self.table.addStagePosition = self.add_stage_position

        self.view.master.tiling_buttons.buttons["tiling"].config(
            command=self.parent_controller.channels_tab_controller.launch_tiling_wizard
        )

        self.view.master.tiling_buttons.buttons["save_data"].config(
            command=self.export_positions
        )

        self.view.master.tiling_buttons.buttons["load_data"].config(
            command=self.load_positions
        )

        self.view.master.tiling_buttons.buttons["eliminate_tiles"].config(
            command=self.eliminate_tiles
        )

    def eliminate_tiles(self):
        """Eliminate tiles that do not contain tissue."""

        self.parent_controller.execute("eliminate_tiles")

    def set_positions(self, positions):
        """Set positions to multi-position's table

        Parameters
        ----------
        positions : dict
            positions to be set

        Example
        -------
        >>>    positions = {
        >>>    0: {'x': 0, 'y': 0, 'z': 0, 'theta': 0, 'f': 0},
        >>>    1: {'x': 1, 'y': 1, 'z': 1, 'theta': 1, 'f': 1},
        >>>    2: {'x': 2, 'y': 2, 'z': 2, 'theta': 2, 'f': 2}}
        >>>    set_positions(positions)
        """
        axis_dict = {"x": "X", "y": "Y", "z": "Z", "theta": "R", "f": "F"}
        data = {}

        for name in axis_dict:
            data[axis_dict[name]] = list(pos[name] for pos in positions)
        self.table.model.df = pd.DataFrame(data)
        self.table.redraw()
        self.show_verbose_info("loaded new positions")

    def get_positions(self):
        """Return all positions from the Multi-Position Acquisition Interface.

        Returns
        -------
        dict
            positions in the format of {index: {axis: value}}

        Example
        -------
        >>> get_positions()
        """
        axis_dict = {"X": "x", "Y": "y", "Z": "z", "R": "theta", "F": "f"}
        positions = []
        rows = self.table.model.df.shape[0]
        for i in range(rows):
            temp = list(self.table.model.df.iloc[i])
            if (
                len(
                    list(filter(lambda v: type(v) == float and not math.isnan(v), temp))
                )
                == 5
            ):
                temp = dict(self.table.model.df.iloc[i])
                positions.append({})
                for k in axis_dict:
                    positions[i][axis_dict[k]] = temp[k]
        return positions

    def handle_double_click(self, event):
        """Move to a position within the Multi-Position Acquisition Interface.

        When double-clicked the row head, it will call the parent/central controller
        to move stage and update stage view

        Parameters
        ----------
        event : tkinter event
            event that triggers the function
        """
        rowclicked = self.table.get_row_clicked(event)
        df = self.table.model.df
        temp = list(df.loc[rowclicked])
        # validate position
        if list(filter(lambda v: type(v) != int and type(v) != float, temp)):
            #  TODO: show error: position is invalid
            print("position is invalid")
            logger.info("position is invalid")
            return
        position = {
            "x": temp[0],
            "y": temp[1],
            "z": temp[2],
            "theta": temp[3],
            "f": temp[4],
        }
        self.parent_controller.execute("move_stage_and_update_info", position)
        self.show_verbose_info("move stage to", position)

    def get_position_num(self):
        """Return the number of positions in the Multi-Position Acquisition Interface.

        Returns
        -------
        int
            number of positions

        Example
        -------
        >>> get_position_num()
        """
        return self.table.model.df.shape[0]

    def load_positions(self):
        """Load a csv file.

        The valid csv file should contain the line of headers ['X', 'Y', 'Z', 'R', 'F']

        Example
        -------
        >>> load_positions()
        """
        filename = filedialog.askopenfilenames(
            defaultextension=".csv",
            filetypes=(("CSV files", "*.csv"), ("Text files", "*.txt")),
        )
        if not filename:
            return
        df = pd.read_csv(filename[0])
        # validate the csv file
        df.columns = map(lambda v: v.upper(), df.columns)
        cmp_header = df.columns == ["X", "Y", "Z", "R", "F"]
        if not cmp_header.all():
            messagebox.showwarning(
                title="Warning",
                message="The csv file isn't right, it should contain [X, Y, Z, R, F]"
            )
            logger.info("The csv file isn't right, it should contain [X, Y, Z, R, F]")
            return
        model = TableModel(dataframe=df)
        self.table.updateModel(model)
        try:
            self.table.redraw()
        except KeyError:
            pass
        self.show_verbose_info("loaded csv file", filename)

    def export_positions(self):
        """Export the positions in the Multi-Position Acquisition Interface to a
        csv file.

        This function opens a dialog that let the user input a filename
        Then, it will export positions to that csv file

        Example
        -------
        >>> export_positions()
        """
        filename = filedialog.asksaveasfilename(
            defaultextension=".csv",
            filetypes=(("CSV file", "*.csv"), ("Text file", "*.txt")),
        )
        if not filename:
            return
        self.table.model.df.to_csv(filename, index=False)
        self.show_verbose_info("exporting csv file", filename)

    def move_to_position(self):
        """Move to a position within the Multi-Position Acquisition Interface.

        Example
        -------
        >>> move_to_position()
        """
        event = type("MyEvent", (object,), {})
        event.x, event.y = 0, 0
        self.handle_double_click(event)

    def insert_row_func(self):
        """Insert a row in the Multi-Position Acquisition Interface.

        Example
        -------
        >>> insert_row_func()
        """
        self.table.model.addRow(self.table.currentrow)
        self.table.update_rowcolors()
        self.table.redraw()
        self.table.tableChanged()
        self.show_verbose_info("insert a row before current row")

    def add_stage_position(self):
        """Add the current stage position to the Multi-Position Acquisition Interface.

        This function will get the stage's current position,
        Then add it to position list

        Example
        -------
        >>> add_stage_position()
        """
        position = self.parent_controller.execute("get_stage_position")
        self.append_position(position)

    def append_position(self, position):
        """Append a position to the Multi-Position Acquisition Interface.

        Parameters
        ----------
        position : dict
            position in the format of {axis: value}

        Example
        -------
        >>> append_position(position)
        """
        temp = list(map(lambda k: position[k], position))
        self.table.model.df = self.table.model.df.append(
            pd.DataFrame([temp], columns=list("XYZRF")), ignore_index=True
        )
        self.table.currentrow = self.table.model.df.shape[0] - 1
        self.table.update_rowcolors()
        self.table.redraw()
        self.table.tableChanged()
        self.show_verbose_info("add current stage position to position list")

    def remove_positions(self, position_flag_list):
        """Remove positions according to position_flag_list

        Parameters
        ----------
        position_flag_list : list[bool]
            False: the position should be removed
            True: the position should be kept
        """
        positions = self.get_positions()
        l = len(position_flag_list)  # noqa
        new_positions = [
            p for i, p in enumerate(positions) if (i >= l or position_flag_list[i])
        ]
        self.set_positions(new_positions)



# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only
# (subject to the limitations in the disclaimer below)
# provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Standard Library Imports
import os
import sys
import time
import shutil
import platform
from pathlib import Path
from os.path import isfile
from multiprocessing.managers import ListProxy, DictProxy

# Third Party Imports
import yaml

# Local Imports


def get_navigate_path():
    """Establish a program home directory in AppData/Local/.navigate for Windows
    or ~/.navigate for Mac and Linux.

    Returns
    -------
    str
        Path to Navigate home directory.

    Examples
    --------
    >>> get_navigate_path()
    'C:\\Users\\username\\AppData\\Local\\.navigate'
    """
    if platform.system() == "Windows":
        base_directory = os.getenv("LOCALAPPDATA")
    else:
        base_directory = os.getenv("HOME")
    navigate_path = os.path.join(base_directory, ".navigate")

    if not os.path.exists(navigate_path):
        os.mkdir(navigate_path)

    return navigate_path


def get_configuration_paths():
    """Get the paths of the various configuration files used by Navigate.

    Returns
    --------
    configuration_path : str
        Path to file containing global microscope configuration,
        i.e. hardware setup
    experiment_path : str
        Path to file containing per-experiment parameters
    waveform_constants_path : str
        Path to file containing remote focus parameters for all magnifications
    rest_api_path : str
        Path to file containing REST API configuration
    waveform_templates_path : str
        Path to file containing waveform templates
    """
    navigate_directory = get_navigate_path()
    if not os.path.exists(navigate_directory):
        os.mkdir(navigate_directory)
    configuration_directory = Path(os.path.join(navigate_directory, "config"))
    if not os.path.exists(configuration_directory):
        os.mkdir(configuration_directory)

    # Configuration files should be stored in this directory
    configuration_path = Path.joinpath(configuration_directory, "configuration.yaml")
    experiment_path = Path.joinpath(configuration_directory, "experiment.yml")
    waveform_constants_path = Path.joinpath(
        configuration_directory, "waveform_constants.yml"
    )
    rest_api_path = Path.joinpath(configuration_directory, "rest_api_config.yml")
    waveform_templates_path = Path.joinpath(
        configuration_directory, "waveform_templates.yml"
    )

    # If they are not already,
    # copy the default ones that ship with the software to this folder
    if not os.path.exists(configuration_path):
        copy_base_directory = Path(__file__).resolve().parent
        copy_configuration_path = Path.joinpath(
            copy_base_directory, "configuration.yaml"
        )
        shutil.copyfile(copy_configuration_path, configuration_path)

    if not os.path.exists(experiment_path):
        copy_base_directory = Path(__file__).resolve().parent
        copy_experiment_path = Path.joinpath(copy_base_directory, "experiment.yml")
        shutil.copyfile(copy_experiment_path, experiment_path)

    if not os.path.exists(waveform_constants_path):
        copy_base_directory = Path(__file__).resolve().parent
        copy_waveform_constants_path = Path.joinpath(
            copy_base_directory, "waveform_constants.yml"
        )
        shutil.copyfile(copy_waveform_constants_path, waveform_constants_path)

    if not os.path.exists(rest_api_path):
        copy_base_directory = Path(__file__).resolve().parent
        copy_rest_api_path = Path.joinpath(copy_base_directory, "rest_api_config.yml")
        shutil.copyfile(copy_rest_api_path, rest_api_path)

    if not os.path.exists(waveform_templates_path):
        copy_base_directory = Path(__file__).resolve().parent
        copy_waveform_templates_path = Path.joinpath(
            copy_base_directory, "waveform_templates.yml"
        )
        shutil.copyfile(copy_waveform_templates_path, waveform_templates_path)

    return (
        configuration_path,
        experiment_path,
        waveform_constants_path,
        rest_api_path,
        waveform_templates_path,
    )


def load_configs(manager, **kwargs):
    """Load configuration files.

    Parameters
    ----------
    manager : multiprocessing.Manager
        Shares objects (e.g., dict) between processes
    **kwargs
        List of configuration file paths

    Returns
    -------
    config_dict : dict
        Shared dictionary containing amalgamation of input configurations.
    """
    if kwargs == {}:
        print("No files provided to load_yaml_config()")
        sys.exit(1)

    config_dict = manager.dict()
    for config_name, file_path in kwargs.items():
        file_path = Path(file_path)
        assert file_path.exists(), "Configuration File not found: {}".format(file_path)
        with open(file_path) as f:
            try:
                config_data = yaml.load(f, Loader=yaml.FullLoader)
                build_nested_dict(manager, config_dict, config_name, config_data)
            except yaml.YAMLError as yaml_error:
                print(f"Configuration - Yaml Error: {yaml_error}")
                sys.exit(1)

    # return combined dictionary
    return config_dict


def build_nested_dict(manager, parent_dict, key_name, dict_data):
    """Nest dictionaries recursively.

    Parameters
    ----------
    manager : multiprocessing.Manager
        Shares objects (e.g., dict) between processes
    parent_dict : dict
        Dictionary we are adding to
    key_name : str
        Name of dictionary to insert
    dict_data : dict
        Dictionary to insert

    Returns
    -------
    None

    Examples
    --------
    >>> build_nested_dict(manager, parent_dict, key_name, dict_data)
    """
    if type(dict_data) != dict and type(dict_data) != list:
        parent_dict[key_name] = dict_data
        return
    if type(dict_data) == dict:
        d = manager.dict()
        for k in dict_data:
            build_nested_dict(manager, d, k, dict_data[k])
    else:
        d = manager.list()
        for i, v in enumerate(dict_data):
            d.append(None)
            build_nested_dict(manager, d, i, v)
    parent_dict[key_name] = d


def update_config_dict(manager, parent_dict, config_name, new_config) -> bool:
    """Read a new file and update info of the configuration dict.

    Parameters
    ----------
    manager : multiprocessing.Manager
        Shares objects (e.g., dict) between processes
    parent_dict : dict
        Dictionary we are adding to
    config_name : str
        Name of subdictionary to replace
    new_config : dict or str
        Dictionary values or
        yaml file name

    Returns
    -------
    bool
        True or False

    Examples
    --------
    >>> update_config_dict(manager, parent_dict, config_name, new_config)
    """
    if type(new_config) != dict and type(new_config) != list:
        file_path = str(new_config)
        if isfile(file_path) and (
            file_path.endswith(".yml") or file_path.endswith(".yaml")
        ):
            with open(file_path) as f:
                new_config = yaml.load(f, Loader=yaml.FullLoader)
        else:
            return False

    build_nested_dict(manager, parent_dict, config_name, new_config)
    return True


def verify_experiment_config(manager, configuration):
    """Verify configuration (configuration, experiment, waveform_constants) yaml files

    Parameters
    ----------
    manager : multiprocessing.Manager
        Shares objects (e.g., dict) between processes
    configuration: configuration object
        contains all the yaml files
    """
    if type(configuration["experiment"]) is not DictProxy:
        update_config_dict(manager, configuration, "experiment", {})

    # verify/build autofocus parameter setting
    # get autofocus supported devices(stages, remote_focus) from configuration.yaml file
    device_dict = {}
    # get devices: stages, NI remote_focus_device
    device_config = configuration["configuration"]["microscopes"]
    for microscope_name in device_config.keys():
        microscope_config = device_config[microscope_name]
        device_dict[microscope_name] = {}
        if (
            "remote_focus_device" in microscope_config.keys()
            and microscope_config["remote_focus_device"]["hardware"]["type"] == "NI"
        ):
            device_dict[microscope_name]["remote_focus"] = {}
            device_ref = microscope_config["remote_focus_device"]["hardware"]["channel"]
            device_dict[microscope_name]["remote_focus"][device_ref] = True
        if "stage" in microscope_config.keys():
            stages = microscope_config["stage"]["hardware"]
            device_dict[microscope_name]["stage"] = {}
            if type(stages) != ListProxy:
                stages = [stages]
            for stage in stages:
                if not stage["type"].lower().startswith("synthetic"):
                    for axis in stage["axes"]:
                        device_dict[microscope_name]["stage"][axis] = True

    autofocus_sample_setting = {
        "coarse_range": 500,
        "coarse_step_size": 50,
        "coarse_selected": True,
        "fine_range": 50,
        "fine_step_size": 5,
        "fine_selected": True,
        "robust_fit": False,
    }
    if (
        "AutoFocusParameters" not in configuration["experiment"]
        or type(configuration["experiment"]["AutoFocusParameters"]) is not DictProxy
    ):
        update_config_dict(
            manager, configuration["experiment"], "AutoFocusParameters", {}
        )
    autofocus_setting_dict = configuration["experiment"]["AutoFocusParameters"]
    # verify if all the devices have been added to the autofocus parameter dict
    for microscope_name in device_dict:
        if microscope_name not in autofocus_setting_dict.keys():
            update_config_dict(manager, autofocus_setting_dict, microscope_name, {})
        for device in device_dict[microscope_name]:
            if device not in autofocus_setting_dict[microscope_name].keys():
                update_config_dict(
                    manager,
                    autofocus_setting_dict[microscope_name],
                    device,
                    {},
                )
            for device_ref in device_dict[microscope_name][device]:
                if (
                    device_ref
                    not in autofocus_setting_dict[microscope_name][device].keys()
                ):
                    update_config_dict(
                        manager,
                        autofocus_setting_dict[microscope_name][device],
                        device_ref,
                        autofocus_sample_setting,
                    )

    # remove non-consistent autofocus parameter
    for microscope_name in autofocus_setting_dict.keys():
        if microscope_name not in device_dict:
            autofocus_setting_dict.pop(microscope_name)
        else:
            for device in autofocus_setting_dict[microscope_name].keys():
                if device not in device_dict[microscope_name]:
                    autofocus_setting_dict[microscope_name].pop(device)
                else:
                    for device_ref in autofocus_setting_dict[microscope_name][
                        device
                    ].keys():
                        if (
                            device_ref
                            not in autofocus_setting_dict[microscope_name][device]
                        ):
                            autofocus_setting_dict[microscope_name][device].pop(
                                device_ref
                            )

    # saving info
    saving_dict_sample = {
        "root_directory": get_navigate_path(),
        "save_directory": get_navigate_path(),
        "user": "Kevin",
        "tissue": "Lung",
        "celltype": "MV3",
        "label": "GFP",
        "file_type": "TIFF",
        "date": time.strftime("%Y-%m-%d"),
        "solvent": "BABB",
    }
    if (
        "Saving" not in configuration["experiment"]
        or type(configuration["experiment"]["Saving"]) is not DictProxy
    ):
        update_config_dict(
            manager, configuration["experiment"], "Saving", saving_dict_sample
        )
    saving_setting_dict = configuration["experiment"]["Saving"]
    for k in saving_dict_sample:
        if k not in saving_setting_dict:
            saving_setting_dict[k] = saving_dict_sample[k]

    # if root directory/saving direcotry doesn't exist
    if not os.path.exists(saving_setting_dict["root_directory"]):
        saving_setting_dict["root_directory"] = saving_dict_sample["root_directory"]
    if not os.path.exists(saving_setting_dict["save_directory"]):
        saving_setting_dict["save_directory"] = saving_dict_sample["save_directory"]

    # camera parameters
    camera_parameters_dict_sample = {
        "x_pixels": 2048,
        "y_pixels": 2048,
        "img_x_pixels": 2048,
        "img_y_pixels": 2048,
        "sensor_mode": "Normal",
        "readout_direction": "Top-to-Bottom",
        "number_of_pixels": 10,
        "binning": "1x1",
        "frames_to_average": 1,
        "databuffer_size": 100,
    }
    if (
        "CameraParameters" not in configuration["experiment"]
        or type(configuration["experiment"]["CameraParameters"]) is not DictProxy
    ):
        update_config_dict(
            manager,
            configuration["experiment"],
            "CameraParameters",
            camera_parameters_dict_sample,
        )
    camera_setting_dict = configuration["experiment"]["CameraParameters"]
    for k in camera_parameters_dict_sample:
        if k not in camera_setting_dict.keys():
            camera_setting_dict[k] = camera_parameters_dict_sample[k]
    # binning
    if camera_setting_dict["binning"] not in ["1x1", "2x2", "4x4"]:
        camera_setting_dict["binning"] = "1x1"
    # x_pixels and y_pixels
    try:
        camera_setting_dict["x_pixels"] = int(camera_setting_dict["x_pixels"])
    except ValueError:
        camera_setting_dict["x_pixels"] = camera_parameters_dict_sample["x_pixels"]

    try:
        camera_setting_dict["y_pixels"] = int(camera_setting_dict["y_pixels"])
    except ValueError:
        camera_setting_dict["y_pixels"] = camera_parameters_dict_sample["y_pixels"]

    # image width and height
    if camera_setting_dict["x_pixels"] <= 0:
        camera_setting_dict["x_pixels"] = camera_parameters_dict_sample["x_pixels"]
    if camera_setting_dict["y_pixels"] <= 0:
        camera_setting_dict["y_pixels"] = camera_parameters_dict_sample["y_pixels"]
    x_binning = int(camera_setting_dict["binning"][0])
    y_binning = int(camera_setting_dict["binning"][2])
    img_x_pixels = camera_setting_dict["x_pixels"] // x_binning
    img_y_pixels = camera_setting_dict["y_pixels"] // y_binning
    camera_setting_dict["img_x_pixels"] = img_x_pixels
    camera_setting_dict["img_y_pixels"] = img_y_pixels

    # sensor mode
    if camera_setting_dict["sensor_mode"] not in ["Normal", "Light-Sheet"]:
        camera_setting_dict["sensor_mode"] = "Normal"
    if camera_setting_dict["readout_direction"] not in [
        "Top-to-Bottom",
        "Bottom-to-Top",
        "Bidirectional",
        "Rev. Bidirectional"
    ]:
        camera_setting_dict["readout_direction"] = "Top-to-Bottom"

    # databuffer_size, number_of_pixels
    for k in ["databuffer_size", "number_of_pixels", "frames_to_average"]:
        try:
            camera_setting_dict[k] = int(camera_setting_dict[k])
        except ValueError:
            camera_setting_dict[k] = camera_parameters_dict_sample[k]
        if camera_setting_dict[k] < 1:
            camera_setting_dict[k] = camera_parameters_dict_sample[k]

    # stage parameters
    stage_dict_sample = {}
    device_config = configuration["configuration"]["microscopes"]
    for microscope_name in device_config.keys():
        stage_dict_sample[microscope_name] = {}
        for k in ["z_step", "f_step", "theta_step"]:
            stage_dict_sample[microscope_name][k] = int(
                device_config[microscope_name]["stage"].get(k, 30)
            )
        stage_dict_sample[microscope_name]["xy_step"] = min(
            device_config[microscope_name]["stage"].get("x_step", 500),
            device_config[microscope_name]["stage"].get("y_step", 500),
        )

    if (
        "StageParameters" not in configuration["experiment"]
        or type(configuration["experiment"]["StageParameters"]) is not DictProxy
    ):
        update_config_dict(
            manager, configuration["experiment"], "StageParameters", stage_dict_sample
        )
    stage_setting_dict = configuration["experiment"]["StageParameters"]
    if "limits" not in stage_setting_dict.keys():
        stage_setting_dict["limits"] = True
    elif type(stage_setting_dict["limits"]) is not bool:
        stage_setting_dict["limits"] = True

    for microscope_name in stage_dict_sample:
        if (
            microscope_name not in stage_setting_dict.keys()
            or type(stage_setting_dict[microscope_name]) is not DictProxy
        ):
            update_config_dict(
                manager,
                stage_setting_dict,
                microscope_name,
                stage_dict_sample[microscope_name],
            )
        else:
            for k in stage_dict_sample[microscope_name]:
                if k not in stage_setting_dict[microscope_name].keys():
                    stage_setting_dict[microscope_name][k] = stage_dict_sample[
                        microscope_name
                    ][k]
                else:
                    try:
                        stage_setting_dict[microscope_name][k] = int(
                            stage_setting_dict[microscope_name][k]
                        )
                    except ValueError:
                        stage_setting_dict[microscope_name][k] = stage_dict_sample[
                            microscope_name
                        ][k]

    # microscope state parameters
    microscope_name = configuration["configuration"]["microscopes"].keys()[0]
    zoom = configuration["configuration"]["microscopes"][microscope_name]["zoom"][
        "position"
    ].keys()[0]
    microscope_state_dict_sample = {
        "microscope_name": microscope_name,
        "image_mode": "live",
        "zoom": zoom,
        "stack_cycling_mode": "per_stack",
        "start_position": 0.0,
        "end_position": 100.0,
        "step_size": 20.0,
        "number_z_steps": 5,
        "timepoints": 1,
        "stack_pause": 0.0,
        "is_save": False,
        "stack_acq_time": 1.0,
        "timepoint_interval": 0,
        "experiment_duration": 1.03,
        "is_multiposition": False,
        "multiposition_count": 1,
        "selected_channels": 0,
        "stack_z_origin": 0,
        "stack_focus_origin": 0,
        "start_focus": 0.0,
        "end_focus": 0.0,
        "abs_z_start": 0.0,
        "abs_z_end": 100.0,
        "scanrange": 500.0,
        "n_plane": 1.0,
        "offset_start": 0.0,
        "offset_end": 9.8,
        "conpro_cycling_mode": "per_stack",
        "waveform_template": "Default",
    }
    if (
        "MicroscopeState" not in configuration["experiment"]
        or type(configuration["experiment"]["MicroscopeState"]) is not DictProxy
    ):
        update_config_dict(
            manager,
            configuration["experiment"],
            "MicroscopeState",
            microscope_state_dict_sample,
        )
    microscope_setting_dict = configuration["experiment"]["MicroscopeState"]
    for k in microscope_state_dict_sample:
        if k not in microscope_setting_dict.keys():
            microscope_setting_dict[k] = microscope_state_dict_sample[k]
        elif not isinstance(
            microscope_setting_dict[k], type(microscope_state_dict_sample[k])
        ):
            if isinstance(microscope_state_dict_sample[k], float):
                try:
                    microscope_setting_dict[k] = float(microscope_setting_dict[k])
                except ValueError:
                    microscope_setting_dict[k] = microscope_state_dict_sample[k]
            elif isinstance(microscope_state_dict_sample[k], int):
                try:
                    microscope_setting_dict[k] = int(microscope_setting_dict[k])
                except ValueError:
                    microscope_setting_dict[k] = microscope_state_dict_sample[k]
            else:
                microscope_setting_dict[k] = microscope_state_dict_sample[k]

    # verify microscope name
    if (
        microscope_setting_dict["microscope_name"]
        not in configuration["configuration"]["microscopes"].keys()
    ):
        microscope_setting_dict["microscope_name"] = microscope_name
    microscope_name = microscope_setting_dict["microscope_name"]
    # zoom
    if (
        microscope_setting_dict["zoom"]
        not in configuration["configuration"]["microscopes"][microscope_name]["zoom"][
            "position"
        ].keys()
    ):
        microscope_setting_dict["zoom"] = configuration["configuration"]["microscopes"][
            microscope_name
        ]["zoom"]["position"].keys()[0]
    # channels
    if (
        "channels" not in microscope_setting_dict
        or type(microscope_setting_dict["channels"]) is not DictProxy
    ):
        update_config_dict(manager, microscope_setting_dict, "channels", {})
    laser_list = [
        f"{laser['wavelength']}nm"
        for laser in configuration["configuration"]["microscopes"][microscope_name][
            "lasers"
        ]
    ]
    filterwheel_list = list(
        configuration["configuration"]["microscopes"][microscope_name]["filter_wheel"][
            "available_filters"
        ].keys()
    )
    prefix = "channel_"
    channel_nums = configuration["configuration"]["gui"]["channels"]["count"]
    channel_setting_dict = microscope_setting_dict["channels"]
    selected_channel_num = 0
    for channel in channel_setting_dict.keys():
        if not channel.startswith(prefix):
            del channel_setting_dict[channel]
            continue
        channel_id = int(channel[len(prefix) :]) - 1
        if channel_id < 0 or channel_id >= channel_nums:
            del channel_setting_dict[channel]
            continue
        channel_value = channel_setting_dict[channel]
        # make sure channel values are right
        # laser
        if channel_value["laser"] not in laser_list:
            channel_value["laser"] = laser_list[0]
        channel_value["laser_index"] = laser_list.index(channel_value["laser"])
        # filter wheel
        if channel_value["filter"] not in filterwheel_list:
            channel_value["filter"] = filterwheel_list[0]
        channel_value["filter_position"] = filterwheel_list.index(
            channel_value["filter"]
        )
        # is_selected
        if (
            "is_selected" not in channel_value.keys()
            or type(channel_value["is_selected"]) != bool
        ):
            channel_value["is_selected"] = False
        if channel_value["is_selected"]:
            selected_channel_num += 1
        # camera_exposure_time and defoucus should be float
        temp = {
            "laser_power": 20.0,
            "camera_exposure_time": 200.0,
            "interval_time": 0.0,
            "defocus": 0.0,
        }
        for k in temp:
            try:
                channel_value[k] = float(channel_value[k])
            except ValueError:
                channel_value[k] = temp[k]
            if channel_value[k] < 0:
                channel_value[k] = temp[k]

    microscope_setting_dict["selected_channels"] = selected_channel_num

    # MultiPositions
    if (
        "MultiPositions" not in configuration["experiment"]
        or type(configuration["experiment"]["MultiPositions"]) is not ListProxy
    ):
        update_config_dict(manager, configuration["experiment"], "MultiPositions", [])
    position_ids = []
    multipositions = configuration["experiment"]["MultiPositions"]
    for i, position in enumerate(multipositions):
        for axis in ["x", "y", "z", "theta", "f"]:
            try:
                position[axis] = float(position[axis])
            except ValueError:
                position_ids.append(i)
                break
    for idx in position_ids[::-1]:
        del multipositions[idx]
    if len(multipositions) < 1:
        multipositions.append(None)
        update_config_dict(
            manager,
            multipositions,
            0,
            {"x": 10.0, "y": 10.0, "z": 10.0, "f": 10.0, "theta": 10.0},
        )
    microscope_setting_dict["multiposition_count"] = len(multipositions)


def verify_waveform_constants(manager, configuration):
    """Verifies and updates the waveform constants in the configuration dictionary.

    This function checks and ensures that the waveform constants in the given
    configuration dictionary conform to the expected structure. It verifies and
    updates the constants related to remote focus devices, lasers, and galvos
    for multiple microscopes.

    Parameters
    ----------
    manager : multiprocessing.Manager
        Shares objects (e.g., dict) between processes
    configuration : dict
        The configuration dictionary containing waveform constants.

    Note
    ----
        If constants are missing or not in the expected structure, default values
        or empty dictionaries are added as necessary.
    Note
    ----
        Laser and galvo constants are validated and converted to float if possible.
    Note
    ----
        Non-existent microscopes, zoom levels, lasers, and galvos are removed
        from the configuration.

    """
    if type(configuration["waveform_constants"]) is not DictProxy:
        update_config_dict(manager, configuration, "waveform_constants", {})
    waveform_dict = configuration["waveform_constants"]

    # remote_focus_constants
    if (
        "remote_focus_constants" not in waveform_dict.keys()
        or type(waveform_dict["remote_focus_constants"]) is not DictProxy
    ):
        update_config_dict(manager, waveform_dict, "remote_focus_constants", {})

    waveform_dict = waveform_dict["remote_focus_constants"]
    for microscope_name in configuration["configuration"]["microscopes"].keys():
        config_dict = configuration["configuration"]["microscopes"][microscope_name]
        if (
            microscope_name not in waveform_dict.keys()
            or type(waveform_dict[microscope_name]) is not DictProxy
        ):
            update_config_dict(manager, waveform_dict, microscope_name, {})

        # get lasers
        lasers = []
        for laser in config_dict["lasers"]:
            laser_wavelength = f"{laser['wavelength']}nm"
            lasers.append(laser_wavelength)

        for zoom in config_dict["zoom"]["position"].keys():
            if (
                zoom not in waveform_dict[microscope_name].keys()
                or type(waveform_dict[microscope_name][zoom]) is not DictProxy
            ):
                update_config_dict(manager, waveform_dict[microscope_name], zoom, {})

            for laser in lasers:
                if (
                    laser not in waveform_dict[microscope_name][zoom].keys()
                    or type(waveform_dict[microscope_name][zoom][laser])
                    is not DictProxy
                ):
                    update_config_dict(
                        manager,
                        waveform_dict[microscope_name][zoom],
                        laser,
                        {
                            "amplitude": config_dict["remote_focus_device"][
                                "amplitude"
                            ],
                            "offset": config_dict["remote_focus_device"]["offset"],
                            "percent_smoothing": "0",
                            "percent_delay": config_dict["remote_focus_device"][
                                "delay_percent"
                            ],
                        },
                    )
                else:
                    for k in [
                        "amplitude",
                        "offset",
                        "percent_smoothing",
                        "percent_delay",
                    ]:
                        if k not in waveform_dict[microscope_name][zoom][laser].keys():
                            waveform_dict[microscope_name][zoom][laser][
                                k
                            ] = config_dict["remote_focus_device"].get(k, "0")
                        else:
                            try:
                                float(waveform_dict[microscope_name][zoom][laser][k])
                            except ValueError:
                                waveform_dict[microscope_name][zoom][laser][
                                    k
                                ] = config_dict["remote_focus_device"].get(k, "0")

            # delete non-exist lasers
            for k in waveform_dict[microscope_name][zoom].keys():
                if k not in lasers:
                    waveform_dict[microscope_name][zoom].pop(k)

        # delete non-exist zoom
        for k in waveform_dict[microscope_name].keys():
            if k not in config_dict["zoom"]["position"].keys():
                waveform_dict[microscope_name].pop(k)

    # delete non-exist microscope
    for k in waveform_dict.keys():
        if k not in configuration["configuration"]["microscopes"].keys():
            waveform_dict.pop(k)

    # galvo_constants
    waveform_dict = configuration["waveform_constants"]
    if (
        "galvo_constants" not in waveform_dict.keys()
        or type(waveform_dict["galvo_constants"]) is not DictProxy
    ):
        update_config_dict(manager, waveform_dict, "galvo_constants", {})

    waveform_dict = waveform_dict["galvo_constants"]

    # get galvo num
    galvo_num = 0
    for microscope_name in configuration["configuration"]["microscopes"].keys():
        galvo_num = max(
            galvo_num,
            len(
                configuration["configuration"]["microscopes"][microscope_name]["galvo"]
            ),
        )

    for i in range(galvo_num):
        waveform_dict = configuration["waveform_constants"]["galvo_constants"]
        galvo_ref = f"Galvo {i}"
        if (
            galvo_ref not in waveform_dict.keys()
            or type(waveform_dict[galvo_ref]) is not DictProxy
        ):
            update_config_dict(manager, waveform_dict, galvo_ref, {})
        waveform_dict = waveform_dict[galvo_ref]
        for microscope_name in configuration["configuration"]["microscopes"].keys():
            if (
                len(
                    configuration["configuration"]["microscopes"][microscope_name][
                        "galvo"
                    ]
                )
                <= i
            ):
                continue
            config_dict = configuration["configuration"]["microscopes"][microscope_name]
            if (
                microscope_name not in waveform_dict.keys()
                or type(waveform_dict[microscope_name]) is not DictProxy
            ):
                update_config_dict(manager, waveform_dict, microscope_name, {})

            for zoom in config_dict["zoom"]["position"].keys():
                if (
                    zoom not in waveform_dict[microscope_name].keys()
                    or type(waveform_dict[microscope_name][zoom]) is not DictProxy
                ):
                    update_config_dict(
                        manager,
                        waveform_dict[microscope_name],
                        zoom,
                        {
                            "amplitude": "0.11",
                            "offset": config_dict["galvo"][i]["offset"],
                            "frequency": config_dict["galvo"][i]["frequency"],
                        },
                    )
                else:
                    for k in ["amplitude", "offset", "frequency"]:
                        if k not in waveform_dict[microscope_name][zoom].keys():
                            waveform_dict[microscope_name][zoom][k] = config_dict[
                                "galvo"
                            ][i].get(k, "0")
                        else:
                            try:
                                float(waveform_dict[microscope_name][zoom][k])
                            except ValueError:
                                waveform_dict[microscope_name][zoom][k] = config_dict[
                                    "galvo"
                                ][i].get(k, "0")
            # delete non-exist zoom
            for k in waveform_dict[microscope_name].keys():
                if k not in config_dict["zoom"]["position"].keys():
                    waveform_dict[microscope_name].pop(k)
        # delete non-exist microscope
        for k in waveform_dict.keys():
            if k not in configuration["configuration"]["microscopes"].keys():
                waveform_dict.pop(k)

    # other_constants
    waveform_dict = configuration["waveform_constants"]
    if (
        "other_constants" not in waveform_dict.keys()
        or type(waveform_dict["other_constants"]) is not DictProxy
    ):
        update_config_dict(
            manager,
            waveform_dict,
            "other_constants",
            {"remote_focus_settle_duration": "0"},
        )
    if "remote_focus_settle_duration" not in waveform_dict["other_constants"].keys():
        waveform_dict["other_constants"]["remote_focus_settle_duration"] = "0"
    else:
        try:
            float(waveform_dict["other_constants"]["remote_focus_settle_duration"])
        except ValueError:
            waveform_dict["other_constants"]["remote_focus_settle_duration"] = "0"



# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only (subject to the
# limitations in the disclaimer below) provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

# Standard library imports
import time
from functools import reduce
from threading import Lock
import copy

# Third party imports

# Local application imports
from .image_writer import ImageWriter
from navigate.tools.common_functions import VariableWithLock


class ChangeResolution:
    """
    ChangeResolution class for modifying the resolution mode of a microscope.

    This class provides functionality to change the resolution mode of a microscope by
    reconfiguring the microscope settings and updating the active microscope.

    Notes:
    ------
    - This class is used to change the resolution mode of a microscope by updating the
    microscope settings and configuring the active microscope accordingly.

    - The `resolution_mode` parameter specifies the desired resolution mode, and the
    `zoom_value` parameter specifies the zoom value to be set. These parameters can
    be adjusted to modify the microscope's configuration.

    - The `ChangeResolution` class is typically used to adapt the microscope's settings
    for different imaging requirements during microscopy experiments.

    - The resolution change process involves reconfiguring the microscope, updating the
    active microscope instance, and resuming data acquisition.

    - The `config_table` attribute is used to define the configuration for the
    resolution change process, including signal acquisition and cleanup steps.
    """

    def __init__(self, model, resolution_mode="high", zoom_value="N/A"):
        """Initialize the ChangeResolution class.


        Parameters:
        ----------
        model : MicroscopeModel
            The microscope model object used for resolution mode changes.
        resolution_mode : str, optional
            The desired resolution mode to set for the microscope. Default is "high".
        zoom_value : str, optional
            The zoom value to set for the microscope. Default is "N/A".
        """
        #: MicroscopeModel: The microscope model associated with the resolution change.
        self.model = model

        #: dict: A dictionary defining the configuration for the resolution change
        self.config_table = {
            "signal": {"main": self.signal_func, "cleanup": self.cleanup},
            "node": {"device_related": True},
        }

        #: str: The desired resolution mode to set for the microscope.
        self.resolution_mode = resolution_mode

        #: str: The zoom value to set for the microscope.
        self.zoom_value = zoom_value

    def signal_func(self):
        """Perform actions to change the resolution mode and update the active
         microscope.

        This method carries out actions to change the resolution mode of the microscope
         by reconfiguring the microscope settings, updating the active microscope, and
         resuming data acquisition.

        Returns:
        -------
        bool
            A boolean value indicating the success of the resolution change process.
        """
        # pause data thread
        self.model.pause_data_thread()
        # end active microscope
        self.model.active_microscope.end_acquisition()
        # prepare new microscope
        self.model.configuration["experiment"]["MicroscopeState"][
            "microscope_name"
        ] = self.resolution_mode
        self.model.configuration["experiment"]["MicroscopeState"][
            "zoom"
        ] = self.zoom_value
        self.model.change_resolution(self.resolution_mode)
        self.model.logger.debug(f"current resolution is {self.resolution_mode}")
        self.model.logger.debug(
            f"current active microscope is {self.model.active_microscope_name}"
        )
        # prepare active microscope
        waveform_dict = self.model.active_microscope.prepare_acquisition()
        self.model.event_queue.put(("waveform", waveform_dict))
        # resume data thread
        self.model.resume_data_thread()
        return True

    def cleanup(self):
        """Perform cleanup actions if needed.

        This method is responsible for performing cleanup actions if required after the
        resolution change process.
        """
        self.model.resume_data_thread()


class Snap:
    """Snap class for capturing data frames using a microscope.

    This class provides functionality to capture data frames using a microscope
    and log information about the camera and frame IDs.

    Notes:
    ------
    - This class is used to capture data frames using a microscope and log
    relevant information, such as the active camera and frame IDs.

    - The `Snap` class is typically used for capturing individual frames during
    microscopy experiments.

    - The data capture process involves capturing frames and logging
    camera-related information.

    - The `config_table` attribute is used to define the configuration for the
    data capture process, specifically the main data capture function.
    """

    def __init__(self, model):
        """Initialize the Snap class.

        Parameters:
        ----------
        model : MicroscopeModel
            The microscope model object used for data capture.
        """
        #: MicroscopeModel: The microscope model associated with the data capture.
        self.model = model

        #: dict: A dictionary defining the configuration for the data capture process.
        self.config_table = {"data": {"main": self.data_func}}

    def data_func(self, frame_ids):
        """Capture data frames and log camera information.

        This method captures data frames using the microscope and logs information
        about the active camera and the provided frame IDs.

        Parameters:
        ----------
        frame_ids : list
            A list of frame IDs for which data frames should be captured.

        Returns:
        -------
        bool
            A boolean value indicating the success of the data capture process.
        """
        self.model.logger.info(
            f"the camera is:{self.model.active_microscope_name}, {frame_ids}"
        )
        return True


class WaitToContinue:
    """WaitToContinue class for synchronizing signal and data acquisition.

    This feature is used to synchronize signal and data acquisition processes, allowing
    the faster one to wait until the other one ends.

    Notes:
    ------
    - This class is used to synchronize signal and data acquisition processes in a
    controlled manner. It ensures that the faster process waits for the slower one to
    complete, improving synchronization during microscopy experiments.

    - The synchronization process involves using locks to control when the signal and
    data acquisition processes are allowed to proceed.

    - The `config_table` attribute defines the configuration for each stage of the
    synchronization process, including initialization, main execution, and cleanup
    steps.

    - The `first_enter_node` attribute tracks which process (signal or data) enters the
    node first to determine synchronization order.

    - The `WaitToContinue` class helps maintain order and synchronization between signal
    and data acquisition nodes in microscopy experiments.
    """

    def __init__(self, model):
        """Initialize the WaitToContinue class.

        Parameters:
        ----------
        model : MicroscopeModel
            The microscope model object used for synchronization.
        """
        #: MicroscopeModel: The microscope model associated with the synchronization.
        self.model = model

        #: Lock: A lock for the signal acquisition process.
        self.pause_signal_lock = Lock()

        #: Lock: A lock for the data acquisition process.
        self.pause_data_lock = Lock()

        #: VariableWithLock: A variable to track which process enters the node first.
        self.first_enter_node = VariableWithLock(str)

        #: dict: A dictionary defining the configuration for the synchronization
        self.config_table = {
            "signal": {
                "init": self.pre_signal_func,
                "main": self.signal_func,
                "cleanup": self.cleanup,
            },
            "data": {
                "init": self.pre_data_func,
                "main": self.data_func,
                "cleanup": self.cleanup,
            },
        }

    def pre_signal_func(self):
        """Prepare for the signal acquisition stage and synchronize with data
        acquisition.

        This method prepares for the signal acquisition stage and synchronizes with
        the data acquisition process, ensuring that the slower process proceeds first.
        """
        with self.first_enter_node as first_enter_node:
            if first_enter_node.value == "":
                self.model.logger.debug("*** wait to continue enters signal " "first!")
                first_enter_node.value = "signal"
                if not self.pause_signal_lock.locked():
                    self.pause_signal_lock.acquire()
                if self.pause_data_lock.locked():
                    self.pause_data_lock.release()

    def signal_func(self):
        """Synchronize signal acquisition and release locks.

        This method synchronizes the signal acquisition process with data acquisition
        and releases any locks held.

        Returns:
        -------
        bool
           A boolean value indicating the success of the synchronization process.
        """
        self.model.logger.debug(f"--wait to continue: {self.model.frame_id}")
        if self.pause_signal_lock.locked():
            self.pause_signal_lock.acquire()
        elif self.pause_data_lock.locked():
            self.pause_data_lock.release()
        self.first_enter_node.value = ""
        self.model.logger.debug(f"--wait to continue is done!: {self.model.frame_id}")
        return True

    def pre_data_func(self):
        """Prepare for the data acquisition stage and synchronize with signal
        acquisition.

        This method prepares for the data acquisition stage and synchronizes with the
        signal acquisition process, ensuring that the slower process proceeds first.
        """
        with self.first_enter_node as first_enter_node:
            if first_enter_node.value == "":
                self.model.logger.debug(
                    "*** wait to continue enters data " "node first!"
                )
                first_enter_node.value = "data"
                if not self.pause_data_lock.locked():
                    self.pause_data_lock.acquire()
                if self.pause_signal_lock.locked():
                    self.pause_signal_lock.release()

    def data_func(self, frame_ids):
        """Synchronize data acquisition and release locks.

        This method synchronizes the data acquisition process with signal acquisition
        and releases any locks held.

        Parameters:
        ----------
        frame_ids : list
            A list of frame IDs for which data acquisition should be performed.

        Returns:
        -------
        bool
            A boolean value indicating the success of the synchronization process.
        """
        self.model.logger.debug(f"**wait to continue? {frame_ids}")
        if self.pause_data_lock.locked():
            self.pause_data_lock.acquire()
        elif self.pause_signal_lock.locked():
            self.pause_signal_lock.release()
        self.first_enter_node.value = ""
        self.model.logger.debug(f"**wait to continue is done! {frame_ids}")
        return True

    def cleanup(self):
        """Release any remaining locks during cleanup.

        This method releases any locks that may still be held during cleanup.
        """
        if self.pause_signal_lock.locked():
            self.pause_signal_lock.release()
        if self.pause_data_lock.locked():
            self.pause_data_lock.release()


class LoopByCount:
    """LoopByCount class for controlling signal and data acquisition loops.

    This class provides functionality to control signal and data acquisition loops by
    specifying the number of steps or frames to execute.

    Notes:
    ------
    - This class is used to control signal and data acquisition loops by specifying the
    number of steps or frames to execute. It allows for flexible control of the
    acquisition process.

    - The `steps` parameter can be an integer specifying the number of steps/frames
    directly or a string representing a configuration reference to determine the loop
    count dynamically.

    - The loop control process involves tracking the remaining steps/frames and deciding
     whether to continue the loop or exit based on the remaining count.

    - The `LoopByCount` class is useful for controlling the number of acquisitions
    during microscopy experiments, either by specifying a fixed count or by
    dynamically determining it from configuration references.
    """

    def __init__(self, model, steps=1):
        """Initialize the LoopByCount class.

        Parameters:
        ----------
        model : MicroscopeModel
            The microscope model object used for loop control.
        steps : int or str, optional
            The number of steps or a configuration reference to determine the loop
            count. Default is 1.
        """
        #: MicroscopeModel: The microscope model associated with the loop control.
        self.model = model

        #: bool: A boolean value indicating whether to step by frame or by step.
        self.step_by_frame = True

        #: int: The remaining number of steps or frames.
        self.steps = steps
        if type(steps) is str:
            self.step_by_frame = False
            try:
                parameters = steps.split(".")
                config_ref = reduce((lambda pre, n: f"{pre}['{n}']"), parameters, "")
                exec(f"self.steps = int(self.model.configuration{config_ref})")
            except:  # noqa
                self.steps = 1

        #: int: The remaining number of steps.
        self.signals = self.steps

        #: int: The remaining number of frames.
        self.data_frames = self.steps

        #: dict: A dictionary defining the configuration for the loop control process.
        self.config_table = {
            "signal": {"main": self.signal_func},
            "data": {"main": self.data_func},
        }

    def signal_func(self):
        """Control the signal acquisition loop and update the remaining steps.

        This method controls the signal acquisition loop by decrementing the remaining
        steps. It determines whether to continue the loop or exit based on the
        remaining count.

        Returns:
        -------
        bool
            A boolean value indicating whether to continue the loop.
        """
        self.signals -= 1
        if self.signals <= 0:
            self.signals = self.steps
            return False
        return True

    def data_func(self, frame_ids):
        """Control the data acquisition loop and update the remaining frames or steps.

        This method controls the data acquisition loop by decrementing the remaining
        frames or steps. It determines whether to continue the loop or exit based on
        the remaining count.

        Parameters:
        ----------
        frame_ids : list
            A list of frame IDs for which data acquisition should be performed.

        Returns:
        -------
        bool
            A boolean value indicating whether to continue the loop.
        """
        if self.step_by_frame:
            self.data_frames -= len(frame_ids)
        else:
            self.data_frames -= 1
        if self.data_frames <= 0:
            self.data_frames = self.steps
            return False
        return True


class PrepareNextChannel:
    """PrepareNextChannel class for preparing microscopes for the next imaging channel.

    This class provides functionality to prepare multiple microscopes, including virtual
    microscopes and the primary microscope, for the next imaging channel during
    microscopy experiments.

    Notes:
    ------
    - This class is used to prepare multiple microscopes for the next imaging
    channel, ensuring that both virtual microscopes and the primary microscope are
    ready for the next step in microscopy experiments.

    - The `PrepareNextChannel` class is typically used to manage the preparation of
    microscopes before transitioning to a new imaging channel.

    - The channel preparation process involves calling the `prepare_next_channel()`
    method for each virtual microscope and the active microscope.

    - The `config_table` attribute is used to define the configuration for the
    channel preparation process, specifically the main preparation function.
    """

    def __init__(self, model):
        """Initialize the PrepareNextChannel class.

        Parameters:
        ----------
        model : MicroscopeModel
            The microscope model object used for channel preparation.
        """
        #: MicroscopeModel: Microscope model associated with the channel preparation.
        self.model = model

        #: dict: A dictionary defining the configuration for the channel preparation
        self.config_table = {"signal": {"main": self.signal_func}}

    def signal_func(self):
        """Prepare virtual and active microscopes for the next imaging channel.

        This method prepares virtual microscopes, if any, followed by the active
        microscope for the next imaging channel.

        Returns:
        -------
        bool
            A boolean value indicating the success of the channel preparation process.
        """
        # prepare virtual microscopes before the primary microscope
        for microscope_name in self.model.virtual_microscopes:
            self.model.virtual_microscopes[microscope_name].prepare_next_channel()

        self.model.active_microscope.prepare_next_channel()

        return True


class MoveToNextPositionInMultiPositionTable:
    """MoveToNextPositionInMultiPositionTable class for advancing in a multi-position
    table.

    This class provides functionality to move to the next position in a multi-position
    table and control the data thread accordingly.

    Notes:
    ------
    - This class is used to advance to the next position in a multi-position table,
    controlling the data thread based on stage distance thresholds.

    - The `MoveToNextPositionInMultiPositionTable` class is typically used to automate
     position changes during microscopy experiments, ensuring proper data thread
     management.

    - The position control process involves moving to the next position in the table,
    pausing the data thread if necessary, and resuming it after the movement.

    - The `config_table` attribute defines the configuration for the position control
    process, including signal acquisition and cleanup steps.
    """

    def __init__(self, model):
        """Initialize the MoveToNextPositionInMultiPositionTable class.

        Parameters:
        ----------
        model : MicroscopeModel
            The microscope model object used for position control.
        """
        #: MicroscopeModel: The microscope model associated with position control.
        self.model = model

        #: dict: A dictionary defining the configuration for the position control
        self.config_table = {
            "signal": {
                "main": self.signal_func,
                "cleanup": self.cleanup,
            },
            "node": {"device_related": True},
        }

        #: int: The current index of the position being acquired in the multi-position
        self.pre_z = None

        #: int: The current index of the position being acquired in the multi-position
        self.current_idx = 0

        #: dict: A dictionary defining the configuration for the position control
        self.multiposition_table = self.model.configuration["experiment"][
            "MultiPositions"
        ]

        #: int: The total number of positions in the multi-position table.
        self.position_count = self.model.configuration["experiment"]["MicroscopeState"][
            "multiposition_count"
        ]

        #: int: The stage distance threshold for pausing the data thread.
        self.stage_distance_threshold = 1000

    def signal_func(self):
        """Move to the next position in the multi-position table and control the data
        thread.

        This method advances to the next position in the multi-position table,
        controls the data thread based on stage distance thresholds, and updates
        position-related information.

        Returns:
        -------
        bool
            A boolean value indicating whether to continue the position control process.
        """
        self.model.logger.debug(
            f"multi-position current idx: {self.current_idx}, {self.position_count}"
        )
        if self.current_idx >= self.position_count:
            return False
        pos_dict = self.multiposition_table[self.current_idx]
        # pause data thread if necessary
        if self.current_idx == 0:
            temp = self.model.get_stage_position()
            pre_stage_pos = dict(
                map(
                    lambda k: (k, temp[f"{k}_pos"]),
                    ["x", "y", "z", "f", "theta"],
                )
            )
        else:
            pre_stage_pos = self.multiposition_table[self.current_idx - 1]
        delta_x = abs(pos_dict["x"] - pre_stage_pos["x"])
        delta_y = abs(pos_dict["y"] - pre_stage_pos["y"])
        delta_z = abs(pos_dict["z"] - pre_stage_pos["z"])
        delta_f = abs(pos_dict["f"] - pre_stage_pos["f"])
        should_pause_data_thread = any(
            distance > self.stage_distance_threshold
            for distance in [delta_x, delta_y, delta_z, delta_f]
        )
        if should_pause_data_thread:
            self.model.pause_data_thread()

        self.current_idx += 1
        abs_pos_dict = dict(map(lambda k: (f"{k}_abs", pos_dict[k]), pos_dict.keys()))
        self.model.logger.debug(f"MoveToNextPositionInMultiPosition: " f"{pos_dict}")
        self.model.move_stage(abs_pos_dict, wait_until_done=True)

        self.model.logger.debug("MoveToNextPositionInMultiPosition: move done")
        # resume data thread
        if should_pause_data_thread:
            self.model.resume_data_thread()
        self.model.active_microscope.central_focus = None
        if self.pre_z != pos_dict["z"]:
            self.pre_z = pos_dict["z"]
            return True

    def cleanup(self):
        """Cleanup method to resume the data thread.

        This method is responsible for resuming the data thread after position control.
        """
        self.model.resume_data_thread()


class StackPause:
    """StackPause class for pausing stack acquisition.

    This class provides functionality to pause stack acquisition for a specified
    number of timepoints or based on a defined pause time. It manages the data thread
    accordingly.

    Notes:
    ------
    - This class is used to pause stack acquisition for a specified number of timepoints
    or based on a defined pause time during microscopy experiments.

    - The `StackPause` class allows for flexible control of stack acquisition pauses,
    ensuring synchronization with data acquisition.

    - The stack pause control process involves managing the data thread, calculating
    pause times, and handling stack acquisition pauses.

    - The `config_table` attribute defines the configuration for the stack pause control
    process, specifically the main pause function.
    """

    def __init__(self, model, pause_num="experiment.MicroscopeState.timepoints"):
        """Initialize the StackPause class.

        Parameters:
        ----------
        model : MicroscopeModel
            The microscope model object used for stack acquisition control.

        pause_num : int or str, optional
            The number of timepoints to pause stack acquisition or a configuration
            reference to determine the pause count dynamically. Default is
            "experiment.MicroscopeState.timepoints".
        """
        self.model = model
        self.pause_num = pause_num
        if type(pause_num) is str:
            try:
                parameters = pause_num.split(".")
                config_ref = reduce((lambda pre, n: f"{pre}['{n}']"), parameters, "")
                exec(f"self.pause_num = int(self.model.configuration{config_ref})")
            except:  # noqa
                self.pause_num = 1
        self.config_table = {"signal": {"main": self.signal_func}}

    def signal_func(self):
        """Pause stack acquisition based on timepoints or pause time and manage the
        data thread.

        This method pauses stack acquisition based on the remaining timepoints or
        defined pause time. It manages the data thread accordingly during the pause.
        """
        self.pause_num -= 1
        if self.pause_num <= 0:
            return
        pause_time = float(
            self.model.configuration["experiment"]["MicroscopeState"]["stack_pause"]
        )
        if pause_time <= 0:
            return
        current_channel = f"channel_{self.model.active_microscope.current_channel}"
        current_exposure_time = (
            float(
                self.model.configuration["experiment"]["MicroscopeState"]["channels"][
                    current_channel
                ]["camera_exposure_time"]
            )
            / 1000.0
        )
        if pause_time < 5 * current_exposure_time:
            time.sleep(pause_time)
        else:
            self.model.pause_data_thread()
            pause_time -= 2 * current_exposure_time
            while pause_time > 0:
                pt = min(pause_time, 0.1)
                time.sleep(pt)
                if self.model.stop_acquisition:
                    self.model.resume_data_thread()
                    return
                pause_time -= 0.1
            self.model.resume_data_thread()


class ZStackAcquisition:
    """ZStackAcquisition class for controlling z-stack acquisition in microscopy.

    This class provides functionality to control z-stack acquisition, including managing
    z and focus positions, acquiring image data, and handling multi-channel
    acquisitions.

    Notes:
    ------
    - This class is used to control z-stack acquisition during microscopy experiments,
    allowing for position cycling, multi-channel acquisitions, and image saving.

    - The z-stack acquisition process involves initializing parameters, controlling
    position and focus movements, handling data acquisition, and managing data threads.

    - The `config_table` attribute defines the configuration for the z-stack acquisition
    process, including signal acquisition, data handling, and node type.
    """

    def __init__(
        self, model, get_origin=False, saving_flag=False, saving_dir="z-stack"
    ):
        """Initialize the ZStackAcquisition class.

        Parameters:
        ----------
        model : MicroscopeModel
            The microscope model object used for z-stack acquisition control.
        get_origin : bool, optional
            Flag to determine whether to get the z and focus origin positions.
            Default is False.
        saving_flag : bool, optional
            Flag to enable image saving during z-stack acquisition. Default is False.
        saving_dir : str, optional
            The sub-directory for saving z-stack images. Default is "z-stack".

        """
        #: MicroscopeModel: The microscope model associated with z-stack acquisition.
        self.model = model

        #: bool: Flag to determine whether to get the z and focus origin positions.
        self.get_origin = get_origin

        #: int: The number of z steps in the z-stack.
        self.number_z_steps = 0

        #: float: The start z position for the z-stack.
        self.start_z_position = 0

        #: float: The start focus position for the z-stack.
        self.start_focus = 0

        #: float: The z step size for the z-stack.
        self.z_step_size = 0

        #: float: The z stack distance for the z-stack.
        self.focus_step_size = 0

        #: dict: A dictionary defining the multi-position table for z-stack acquisition.
        self.positions = {}

        #: int: The current index of the position being acquired in the multi-position
        self.current_position_idx = 0

        #: int: The z position of the channel being acquired in the z-stack
        self.current_z_position = 0

        #: int: The f position of the channel being acquired in the z-stack
        self.current_focus_position = 0

        #: bool: Flag to determine whether to move to a new position
        self.need_to_move_new_position = True

        #: bool: Flag to determine whether to move the z position
        self.need_to_move_z_position = True

        #: int: The number of times the z position has been moved
        self.z_position_moved_time = 0

        #: dict: A dictionary defining the defocus values between channels
        self.defocus = None

        #: str: The stack cycling mode for z-stack acquisition.
        self.stack_cycling_mode = "per_stack"

        #: int: The number of channels in the z-stack.
        self.channels = 1

        #: ImageWriter: An image writer object for saving z-stack images.
        self.image_writer = None
        if saving_flag:
            self.image_writer = ImageWriter(model, sub_dir=saving_dir)

        #: dict: A dictionary defining the configuration for the z-stack acquisition
        self.config_table = {
            "signal": {
                "init": self.pre_signal_func,
                "main": self.signal_func,
                "end": self.signal_end,
            },
            "data": {
                "init": self.pre_data_func,
                "main": self.in_data_func,
                "end": self.end_data_func,
                "cleanup": self.cleanup_data_func,
            },
            "node": {"node_type": "multi-step", "device_related": True},
        }

    def pre_signal_func(self):
        """Initialize z-stack acquisition parameters before the signal stage.

        This method initializes z-stack acquisition parameters, including position,
        focus, and data thread management, before the signal stage.
        """
        microscope_state = self.model.configuration["experiment"]["MicroscopeState"]

        self.stack_cycling_mode = microscope_state["stack_cycling_mode"]

        # get available channels
        self.channels = microscope_state["selected_channels"]
        #: int: The current channel being acquired in the z-stack
        self.current_channel_in_list = 0

        self.number_z_steps = int(microscope_state["number_z_steps"])
        self.start_z_position = float(microscope_state["start_position"])
        # end_z_position = float(microscope_state["end_position"])
        self.z_step_size = float(microscope_state["step_size"])
        #: float: The z stack distance for the z-stack.
        self.z_stack_distance = abs(
            self.start_z_position - float(microscope_state["end_position"])
        )

        self.start_focus = float(microscope_state["start_focus"])
        end_focus = float(microscope_state["end_focus"])
        self.focus_step_size = (end_focus - self.start_focus) / self.number_z_steps
        #: float: The focus stack distance for the z-stack.
        self.f_stack_distance = abs(end_focus - self.start_focus)

        # restore z, f
        pos_dict = self.model.get_stage_position()
        self.model.logger.debug(f"**** ZStack get stage position: {pos_dict}")
        #: float: The z position of the channel being acquired in the z-stack
        self.restore_z = pos_dict["z_pos"]
        #: float: The f position of the channel being acquired in the z-stack
        self.restore_f = pos_dict["f_pos"]

        if bool(microscope_state["is_multiposition"]):
            self.positions = self.model.configuration["experiment"]["MultiPositions"]
        else:
            self.positions = [
                {
                    "x": float(pos_dict["x_pos"]),
                    "y": float(pos_dict["y_pos"]),
                    "z": float(
                        microscope_state.get(
                            "stack_z_origin",
                            pos_dict["z_pos"],
                        )
                        if not self.get_origin
                        else pos_dict["z_pos"]
                    ),
                    "theta": float(pos_dict["theta_pos"]),
                    "f": float(
                        microscope_state.get(
                            "stack_focus_origin",
                            pos_dict["f_pos"],
                        )
                        if not self.get_origin
                        else pos_dict["f_pos"]
                    ),
                }
            ]

        # Setup next channel down here, to ensure defocus isn't merged into
        # restore f_pos, positions
        self.model.active_microscope.central_focus = None
        self.model.active_microscope.current_channel = 0
        self.model.active_microscope.prepare_next_channel()

        self.model.logger.debug(
            f"*** ZStack pre_signal_func: {self.positions}, {self.start_focus}, "
            f"{self.start_z_position}"
        )
        self.current_position_idx = 0
        self.z_position_moved_time = 0
        self.need_to_move_new_position = True
        self.need_to_move_z_position = True
        #: bool: Flag to determine whether to pause the data thread.
        self.should_pause_data_thread = False
        # TODO: distance > 1000 should not be hardcoded and somehow related to
        #  different kinds of stage devices.
        self.stage_distance_threshold = 1000

        self.defocus = [
            v["defocus"]
            for v in microscope_state["channels"].values()
            if v["is_selected"]
        ]

    def signal_func(self):
        """Control z-stack acquisition, move positions, and manage data threads.

        This method controls the z-stack acquisition process, including moving positions
        and focus, managing data threads, and handling data acquisition during the
        signal stage.

        Returns:
        -------
        bool
            A boolean value indicating whether to continue the z-stack acquisition
            process.
        """
        if self.model.stop_acquisition:
            return False
        data_thread_is_paused = False
        # move stage X, Y, Theta
        if self.need_to_move_new_position:
            self.need_to_move_new_position = False

            # calculate first z, f position
            self.current_z_position = (
                self.start_z_position + self.positions[self.current_position_idx]["z"]
            )
            self.current_focus_position = (
                self.start_focus + self.positions[self.current_position_idx]["f"]
            )
            if self.defocus is not None:
                self.current_focus_position += self.defocus[
                    self.current_channel_in_list
                ]

            # calculate delta_x, delta_y
            # TODO: Here.
            pos_dict = dict(
                map(
                    lambda ax: (
                        f"{ax}_abs",
                        self.positions[self.current_position_idx][ax],
                    ),
                    ["x", "y", "theta"],
                )
            )

            if self.current_position_idx > 0:
                delta_x = (
                    self.positions[self.current_position_idx]["x"]
                    - self.positions[self.current_position_idx - 1]["x"]
                )
                delta_y = (
                    self.positions[self.current_position_idx]["y"]
                    - self.positions[self.current_position_idx - 1]["y"]
                )
                delta_z = (
                    self.positions[self.current_position_idx]["z"]
                    - self.positions[self.current_position_idx - 1]["z"]
                    + self.z_stack_distance
                )
                delta_f = (
                    self.positions[self.current_position_idx]["f"]
                    - self.positions[self.current_position_idx - 1]["f"]
                    + self.f_stack_distance
                )
            else:
                delta_x = 0
                delta_y = 0
                delta_z = 0
                delta_f = 0

            # displacement = [delta_z, delta_f, delta_x, delta_y]
            # Check the distance between current position and previous position,
            # if it is too far, then we can call self.model.pause_data_thread() and
            # self.model.resume_data_thread() after the stage has completed the move
            # to the next position.

            self.should_pause_data_thread = any(
                distance > self.stage_distance_threshold
                for distance in [delta_x, delta_y, delta_z, delta_f]
            )
            if self.should_pause_data_thread:
                self.model.pause_data_thread()
                data_thread_is_paused = True

            self.model.move_stage(pos_dict, wait_until_done=True)
            self.model.logger.debug(f"*** ZStack move stage: {pos_dict}")

        if self.need_to_move_z_position:
            # move z, f
            # self.model.pause_data_thread()

            self.model.logger.debug(
                f"*** Zstack move stage: (z: {self.current_z_position}), "
                f"(f: {self.current_focus_position})"
            )

            if self.should_pause_data_thread and not data_thread_is_paused:
                self.model.pause_data_thread()

            self.model.move_stage(
                {
                    "z_abs": self.current_z_position,
                    "f_abs": self.current_focus_position,
                },
                wait_until_done=True,
            )

        if self.should_pause_data_thread:
            self.model.resume_data_thread()
            self.should_pause_data_thread = False

        return True

    def signal_end(self):
        """Handle the end of the signal stage and position cycling.

        This method handles the end of the signal stage, including position cycling and
        channel updates for multi-channel acquisitions.

        Returns:
        -------
        bool
            A boolean value indicating whether to end the current node.
        """

        # end this node
        if self.model.stop_acquisition:
            return True

        if self.stack_cycling_mode != "per_stack":
            # update channel for each z position in 'per_slice'
            if self.defocus is not None:
                self.current_focus_position -= self.defocus[
                    self.current_channel_in_list
                ]
            self.update_channel()
            self.need_to_move_z_position = self.current_channel_in_list == 0

        # in 'per_slice', move to next z position if all the channels have been acquired
        if self.need_to_move_z_position:
            # next z, f position
            self.current_z_position += self.z_step_size
            self.current_focus_position += self.focus_step_size

            # update z position moved time
            self.z_position_moved_time += 1

        # decide whether to move X,Y,Theta
        if self.z_position_moved_time >= self.number_z_steps:
            self.z_position_moved_time = 0
            # calculate first z, f position
            self.current_z_position = (
                self.start_z_position + self.positions[self.current_position_idx]["z"]
            )
            self.current_focus_position = (
                self.start_focus + self.positions[self.current_position_idx]["f"]
            )
            if (
                self.z_stack_distance > self.stage_distance_threshold
                or self.f_stack_distance > self.stage_distance_threshold
            ):
                self.should_pause_data_thread = True

            # after running through a z-stack, update channel
            if self.stack_cycling_mode == "per_stack":
                self.update_channel()
                # if run through all the channels, move to next position
                if self.current_channel_in_list == 0:
                    self.need_to_move_new_position = True
            else:
                self.need_to_move_new_position = True

            if self.need_to_move_new_position:
                # move to next position
                self.current_position_idx += 1

        if self.current_position_idx >= len(self.positions):
            self.current_position_idx = 0
            # restore z
            self.model.move_stage(
                {"z_abs": self.restore_z, "f_abs": self.restore_f},
                wait_until_done=False,
            )  # Update position
            return True

        return False

    def update_channel(self):
        """Update the active channel during multi-channel acquisition.

        This method updates the active channel for multi-channel acquisitions, allowing
        cycling through channels.
        """
        self.current_channel_in_list = (
            self.current_channel_in_list + 1
        ) % self.channels
        # not update DAQ tasks if there is a NI Galvo stage
        self.model.active_microscope.prepare_next_channel()
        if self.defocus is not None:
            self.current_focus_position += self.defocus[self.current_channel_in_list]

    def pre_data_func(self):
        """Initialize data-related parameters before data acquisition.

        This method initializes data-related parameters before data acquisition,
        including the count of received and expected frames.
        """

        self.received_frames = 0
        self.total_frames = self.channels * self.number_z_steps * len(self.positions)

    def in_data_func(self, frame_ids):
        """Handle incoming data frames during data acquisition.

        This method handles incoming data frames during data acquisition, updating the
        count of received frames and saving images if enabled.

        Parameters:
        ----------
        frame_ids : list
            A list of frame IDs received during data acquisition.

        """
        self.model.mark_saving_flags(frame_ids)
        self.received_frames += len(frame_ids)
        if self.image_writer is not None:
            self.image_writer.save_image(frame_ids)

    def end_data_func(self):
        """Check if all expected data frames have been received.

        This method checks whether all expected data frames have been received during
        data acquisition.

        Returns:
        -------
        bool
            A boolean value indicating whether all expected data frames have been
            received.
        """

        return self.received_frames >= self.total_frames

    def cleanup_data_func(self):
        """Perform cleanup actions after data acquisition, if image saving is enabled.

        This method performs cleanup actions after data acquisition, such as cleaning up
         image writing, if image saving is enabled.
        """
        if self.image_writer:
            self.image_writer.cleanup()


class FindTissueSimple2D:
    """FindTissueSimple2D class for detecting tissue and gridding out the imaging
    space in  2D.

    This class is responsible for detecting tissue, thresholding, and gridding out the
    space for 2D imaging. It processes acquired frames to determine regions of
    interest (tissue), calculates offsets, and generates grid positions for imaging.
    """

    def __init__(
        self,
        model,
        overlap=0.1,
        target_resolution="Nanoscale",
        target_zoom="N/A",
    ):
        """Initialize the FindTissueSimple2D class.

        Parameters:
        ----------
        model : MicroscopeModel
            The microscope model object used for tissue detection and gridding.
        overlap : float, optional
            The overlap percentage between grid tiles. Default is 0.1 (10%).
        target_resolution : str, optional
            The target resolution for imaging (e.g., "Nanoscale"). Default is
            "Nanoscale".
        target_zoom : str, optional
            The target zoom level for imaging. Default is "N/A".
        """
        #: MicroscopeModel: The microscope model associated with tissue detection and
        self.model = model

        #: dict: A dictionary defining the configuration for tissue detection and
        self.config_table = {"signal": {}, "data": {"main": self.data_func}}

        #: float: The overlap percentage between grid tiles.
        self.overlap = overlap

        #: str: The target resolution for imaging (e.g., "Nanoscale").
        self.target_resolution = target_resolution

        #: str: The target zoom level for imaging.
        self.target_zoom = target_zoom

    def data_func(self, frame_ids):
        """Process acquired frames for tissue detection, thresholding, and grid
        calculation.

        This method processes acquired frames to detect tissue regions, apply
        thresholding, and calculate grid positions for imaging. It performs the
        following steps:
        - Downsamples the acquired image based on desired magnification change.
        - Applies thresholding using Otsu's method.
        - Calculates the bounding box of the tissue region.
        - Computes offsets for gridding based on acquired and target resolutions.
        - Grids out the 2D imaging space based on the specified overlap percentage.

        Parameters:
        ----------
        frame_ids : list
            A list of frame IDs corresponding to acquired frames.
        """

        from skimage import filters
        from skimage.transform import downscale_local_mean
        import numpy as np
        from navigate.tools.multipos_table_tools import (
            compute_tiles_from_bounding_box,
            calc_num_tiles,
        )

        for idx in frame_ids:
            img = self.model.data_buffer[idx]

            # Get current mag
            microscope_name = self.model.configuration["experiment"]["MicroscopeState"][
                "microscope_name"
            ]
            zoom = self.model.configuration["experiment"]["MicroscopeState"]["zoom"]
            curr_pixel_size = self.model.configuration["configuration"]["microscopes"][
                microscope_name
            ]["zoom"][zoom]["pixel_size"]
            # get target pixel size
            pixel_size = self.model.configuration["configuration"]["microscopes"][
                self.target_resolution
            ]["zoom"][self.target_zoom]["pixel_size"]

            # Downsample according to the desired magnification change. Note, we
            # could downsample by whatever we want.
            ds = int(curr_pixel_size / pixel_size)
            ds_img = downscale_local_mean(img, (ds, ds))

            # Threshold
            thresh_img = ds_img > filters.threshold_otsu(img)

            # Find the bounding box
            # In the real-deal, non-transposed image, x increase corresponds to a
            # decrease in row number y increase responds to an increase in row number
            # This means the smallest x coordinate is actually the largest
            x, y = np.where(thresh_img)
            # + 0.5 accounts for center of FOV
            x_start, x_end = -curr_pixel_size * ds * (
                np.max(x) + 0.5
            ), -curr_pixel_size * ds * (np.min(x) + 0.5)
            y_start, y_end = curr_pixel_size * ds * (
                np.min(y) + 0.5
            ), curr_pixel_size * ds * (np.max(y) + 0.5)
            xd, yd = abs(x_start - x_end), y_end - y_start

            # grab z, theta, f starting positions
            z_start = self.model.configuration["experiment"]["StageParameters"]["z"]
            r_start = self.model.configuration["experiment"]["StageParameters"]["theta"]
            if self.target_resolution == "Nanoscale":
                f_start = 0  # very different range of focus values in high-res
            else:
                f_start = self.model.configuration["experiment"]["StageParameters"]["f"]

            # Update x and y start to initialize from the upper-left corner of the
            # current image, since this is how np.where indexed them. The + 0.5 in
            # x_start/y_start calculation shifts their starts back to the center of the
            # field of view.
            curr_fov_x = (
                float(
                    self.model.configuration["experiment"]["CameraParameters"][
                        "x_pixels"
                    ]
                )
                * curr_pixel_size
            )
            curr_fov_y = (
                float(
                    self.model.configuration["experiment"]["CameraParameters"][
                        "y_pixels"
                    ]
                )
                * curr_pixel_size
            )
            x_start += (
                self.model.configuration["experiment"]["StageParameters"]["x"]
                + curr_fov_x / 2
            )
            y_start += (
                self.model.configuration["experiment"]["StageParameters"]["y"]
                - curr_fov_y / 2
            )

            # stage offset
            x_start += float(
                self.model.configuration["configuration"]["microscopes"][
                    self.target_resolution
                ]["stage"]["x_offset"]
            ) - float(
                self.model.configuration["configuration"]["microscopes"][
                    microscope_name
                ]["stage"]["x_offset"]
            )
            y_start += float(
                self.model.configuration["configuration"]["microscopes"][
                    self.target_resolution
                ]["stage"]["y_offset"]
            ) - float(
                self.model.configuration["configuration"]["microscopes"][
                    microscope_name
                ]["stage"]["y_offset"]
            )
            z_start += float(
                self.model.configuration["configuration"]["microscopes"][
                    self.target_resolution
                ]["stage"]["z_offset"]
            ) - float(
                self.model.configuration["configuration"]["microscopes"][
                    microscope_name
                ]["stage"]["z_offset"]
            )
            r_start += float(
                self.model.configuration["configuration"]["microscopes"][
                    self.target_resolution
                ]["stage"]["r_offset"]
            ) - float(
                self.model.configuration["configuration"]["microscopes"][
                    microscope_name
                ]["stage"]["r_offset"]
            )

            # grid out the 2D space
            fov_x = (
                float(
                    self.model.configuration["experiment"]["CameraParameters"][
                        "x_pixels"
                    ]
                )
                * pixel_size
            )
            fov_y = (
                float(
                    self.model.configuration["experiment"]["CameraParameters"][
                        "y_pixels"
                    ]
                )
                * pixel_size
            )
            x_tiles = calc_num_tiles(xd, self.overlap, fov_x)
            y_tiles = calc_num_tiles(yd, self.overlap, fov_y)

            table_values = compute_tiles_from_bounding_box(
                x_start,
                x_tiles,
                fov_x,
                self.overlap,
                y_start,
                y_tiles,
                fov_y,
                self.overlap,
                z_start,
                1,
                0,
                self.overlap,
                r_start,
                1,
                0,
                self.overlap,
                f_start,
                1,
                0,
                self.overlap,
            )

            self.model.event_queue.put(("multiposition", table_values))



# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only
# (subject to the limitations in the disclaimer below)
# provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Standard library imports
import time
import serial
import threading

# Third-party imports
import numpy as np

# Local application imports


class MP285:
    """Sutter MP285 3D Stage API

    The USB interface expects a baud rate of 9600, 8 data bits, 1 stop bits,
    no parity, and no flow control.

    The serial interface expects a baud rate of 9600, 8 data bits, 1 stop bits,
    no parity, and hardware or RTS/CTS flow control.

    Little-endian byte order is used for all commands and responses.

    Each command sequence consists of at least one byte, the first of which is the
    “command byte”. Those commands that have parameters or arguments require a
    sequence of bytes that follow the command byte. No delimiters are used between
    command sequence arguments, and command sequence terminators are used in most cases.

    Most command sequences have a terminator: ASCII CR (Carriage Return; 13 decimal,
    0D hexadecimal) to indicate that the task associated with the command has completed.
    When the controller completes the task associated with a command, it sends ASCII CR
    back to the host computer indicating that it is ready to receive a new command.
    If a command returns data, the last byte returned is the task-completed indicator.
    """

    def __init__(self, com_port, baud_rate, timeout=0.25):
        self.serial = serial.Serial()
        self.serial.port = com_port
        self.serial.baudrate = baud_rate
        self.serial.timeout = timeout
        self.serial.parity = serial.PARITY_NONE
        self.serial.bytesize = serial.EIGHTBITS
        self.serial.stopbits = serial.STOPBITS_ONE
        self.serial.xonxoff = False
        self.serial.rtscts = True

        self.speed = 1000  # None
        self.resolution = "high"  # None
        self.wait_until_done = True

        self.wait_time = 0.002
        self.n_waits = max(int(timeout / self.wait_time), 1)

        # Thread blocking here to prevent calls to get_current_position()
        # while move_to_specified_position is waiting for a response. Serial
        # commands must complete or the MP-285A completely locks up and has
        # to be power cycled.
        self.safe_to_write = threading.Event()
        self.safe_to_write.set()
        self.last_write_time = time.time()

    def safe_write(self, command):
        self.safe_to_write.wait()
        self.safe_to_write.clear()
        curr_time = time.time()
        # Recommended time between commands is 2 ms
        diff_time = curr_time - self.last_write_time
        if diff_time < 0.002:
            time.sleep(0.002 - diff_time + 0.0001)
        self.serial.read_all()
        self.serial.reset_input_buffer()
        self.serial.reset_output_buffer()
        self.serial.write(command)
        self.last_write_time = time.time()

    def connect_to_serial(self):
        try:
            self.serial.open()
        except serial.SerialException as e:
            print("MP285 serial connection failed!")
            raise e

    def disconnect_from_serial(self):
        self.serial.close()

    def flush_buffers(self):
        """Flush Serial I/O Buffers."""
        self.safe_to_write.wait()
        self.safe_to_write.clear()
        self.serial.read_all()
        self.serial.reset_input_buffer()
        self.serial.reset_output_buffer()
        self.safe_to_write.set()

    @staticmethod
    def convert_microsteps_to_microns(microsteps):
        """Converts microsteps to microns

        Parameters
        ----------
        microsteps : float
            Distance in microsteps

        Returns
        ---------
        microns : float
            Distance in microns
        """

        microns = np.multiply(microsteps, 0.04)
        return microns

    @staticmethod
    def convert_microns_to_microsteps(microns):
        """Converts microsteps to microns.

        Parameters
        ----------
        microns : float
            Distance in microns.

        Returns
        -------
        microsteps : float
            Distance in microsteps
        """

        microsteps = np.divide(microns, 0.04)
        return microsteps

    def get_current_position(self):
        """Get the current stage position.

        Gets the stage position. The data returned consists of 13 bytes:
        12 bytes containing X, Y, & Z position values in microsteps (4 bytes each),
        followed with the task-complete indicator (1 byte).

        Returns
        -------
        x_pos : float
            X position in microns
        y_pos : float
            Y position in microns
        z_pos : float
            Z position in microns

        """
        # print("calling get_current_position")
        # self.flush_buffers()
        command = bytes.fromhex("63") + bytes.fromhex("0d")
        self.safe_write(command)
        # position_information = self.serial.read_until(
        #     expected=bytes.fromhex("0d"), size=100
        # )
        # print(f"sending: {command}")
        position_information = b""
        for _ in range(max(self.n_waits, 13)):
            curr_read = self.serial.read(1)
            if curr_read == b"":
                time.sleep(self.wait_time)
            else:
                position_information += curr_read
            if len(position_information) == 13:
                break
        if len(position_information) != 13:
            raise UserWarning(
                f"Encountered response {position_information}. "
                "You need to power cycle the stage."
            )
        self.safe_to_write.set()
        # print(f"received: {position_information}")
        xs = int.from_bytes(position_information[0:4], byteorder="little", signed=True)
        ys = int.from_bytes(position_information[4:8], byteorder="little", signed=True)
        zs = int.from_bytes(position_information[8:12], byteorder="little", signed=True)
        # print(f"converted to microsteps: {xs} {ys} {zs}")
        x_pos = self.convert_microsteps_to_microns(xs)
        y_pos = self.convert_microsteps_to_microns(ys)
        z_pos = self.convert_microsteps_to_microns(zs)
        # print(f"converted to position: {x_pos} {y_pos} {z_pos}")
        return x_pos, y_pos, z_pos

    def move_to_specified_position(self, x_pos, y_pos, z_pos):
        """Move to Specified Position (‘m’) Command

        This command instructs the controller to move all three axes to the position
        specified.

        The command sequence consists of 14 bytes: Command byte followed by three
        sets of four bytes containing position information in microsteps for X, Y,
        and Z, and ending with the terminator.
        Return data consists of 1 byte (task-complete indicator), which occurs after
        the  move is complete.

        Parameters
        ----------
        x_pos : float
            X position in microns
        y_pos : float
            Y position in microns
        z_pos : float
            Z position in microns

        Returns
        -------
        move_complete : bool
            True if move was successful, False if not.
        """
        # print("calling move_to_specified_position")
        # print(f"moving to {x_pos} {y_pos} {z_pos}")
        # Calculate time to move
        # current_x, current_y, current_z = self.get_current_position()
        # delta_x = abs(x_pos - current_x)
        # delta_y = abs(y_pos - current_y)
        # delta_z = abs(z_pos - current_z)
        # max_distance = max(delta_x, delta_y, delta_z)
        # time_to_move = np.clip(max_distance / self.speed, 0.02, 1.0)
        # print(f"time to move: {time_to_move} s")

        # Convert microns to microsteps and create command.
        x_target = int(self.convert_microns_to_microsteps(x_pos))
        y_target = int(self.convert_microns_to_microsteps(y_pos))
        z_target = int(self.convert_microns_to_microsteps(z_pos))
        x_steps = x_target.to_bytes(length=4, byteorder="little", signed=True)
        y_steps = y_target.to_bytes(length=4, byteorder="little", signed=True)
        z_steps = z_target.to_bytes(length=4, byteorder="little", signed=True)

        # Move stage
        move_cmd = (
            bytes.fromhex("6d") + x_steps + y_steps + z_steps + bytes.fromhex("0d")
        )
        # self.flush_buffers()
        self.safe_write(move_cmd)
        # print(f"move command: {move_cmd} wait_until_done {self.wait_until_done}")

        for _ in range(self.n_waits):
            # time.sleep(time_to_move)
            response = self.serial.read(1)
            # print(f"move response: {response}")
            if response == b"":
                time.sleep(self.wait_time)
            elif response == bytes.fromhex("0d"):
                self.safe_to_write.set()
                return True
            else:
                self.safe_to_write.set()
                # self.flush_buffers()
                # print(f"Uh oh: {response}")
                raise UserWarning(
                    f"Encountered response {response}. "
                    "You probably need to power cycle the stage."
                )

        # # time.sleep(time_to_move)
        # self.safe_to_write.set()

        # response = self.serial.read(1)
        # print(f"move response: {response}")
        # if response == bytes.fromhex("0d"):
        #     move_complete = True
        # else:
        #     move_complete = False
        # return move_complete

        self.safe_to_write.set()
        return False

    def set_resolution_and_velocity(self, speed, resolution):
        """Sets the MP-285 stage speed and resolution.

        This command instructs the controller to move all three axes to the position
        specified. The command sequence consists of 4 bytes: Command byte, followed
        by 2 bytes containing resolution and velocity information, and ending with
        the  terminator. Return data consists of 1 byte (task-complete indicator).

        Parameters
        ----------
        speed : int
            Low Resolution = 0-3000 microns/sec
            High Resolution = 0-1310 microns/sec
        resolution : str
            Low - 0.2 microns/microstep = 10 microsteps/step
            High - 0.04 microns/microstep = 50 microsteps/step

        Returns
        -------
        command_complete : bool
            True if command was successful, False if not.
        """
        # print("calling set_resolution_and_velocity")
        # print(f"resolution: {resolution}")
        if resolution == "high":
            resolution_bit = 1
            if speed > 1310:
                speed = 1310
                raise UserWarning(
                    "High resolution mode of Sutter MP285 speed too "
                    "high. Setting to 1310 microns/sec."
                )
        elif resolution == "low":
            resolution_bit = 0
            if speed > 3000:
                speed = 3000
                raise UserWarning(
                    "Low resolution mode of Sutter MP285 speed too "
                    "high. Setting to 3000 microns/sec."
                )
        else:
            raise UserWarning("MP-285 resolution must be 'high' or 'low'")

        speed_and_res = int(resolution_bit * 32768 + speed)
        command = (
            bytes.fromhex("56")
            + speed_and_res.to_bytes(length=2, byteorder="little", signed=False)
            + bytes.fromhex("0d")
        )

        # Write Command and get response
        # self.flush_buffers()
        self.safe_write(command)
        response = self.serial.read(1)
        # print(f"Response {response}")
        if response == bytes.fromhex("0d"):
            self.speed = speed
            self.resolution = resolution
            command_complete = True
        else:
            command_complete = False
        # print(f"Command complete? {command_complete}")
        self.safe_to_write.set()
        return command_complete

    def interrupt_move(self):
        """Interrupt stage movement.

        This command interrupts and stops a move in progress that originally
        initiated by the Move (‘m’) command. The command sequence consists of 1 byte:
        Command byte (no terminator).

        Return data consists of 1 byte if movement is not in progress, or 2 bytes
        ('=' (move-in- progress indicator)and task-complete indicator.

        Returns
        -------
        stage_stopped : bool
            True if move was successful, False if not.
        """
        # print("calling interrupt_move")

        # Send Command
        self.safe_to_write.set()
        # self.flush_buffers()
        self.safe_write(bytes.fromhex("03"))

        # Get Response
        for _ in range(self.n_waits):
            # time.sleep(time_to_move)
            response = self.serial.read(1)
            # print(f"move response: {response}")
            if response == b"":
                time.sleep(self.wait_time)
            elif response == bytes.fromhex("0d"):
                self.safe_to_write.set()
                return True
            elif response == bytes.fromhex("3d"):
                for _ in range(self.n_waits):
                    response2 = self.serial.read(1)
                    if response2 == b"":
                        time.sleep(self.wait_time)
                    elif response2 == bytes.fromhex("0d"):
                        self.safe_to_write.set()
                        return True

        self.safe_to_write.set()
        return False

    def set_absolute_mode(self):
        """Set MP285 to Absolute Position Mode.

        This command sets the nature of the positional values specified with the Move
        (‘m’) command as absolute positions as measured from the point of origin
        (Position 0). The command sequence consists of 2 bytes: Command byte, followed
        by the terminator. Return data consists of 1 byte (task-complete indicator).

        Returns
        -------
        command_complete : bool
            True if command was successful, False if not.

        """
        # print("calling set_absolute_mode")
        # self.flush_buffers()
        abs_cmd = bytes.fromhex("61") + bytes.fromhex("0d")
        self.safe_write(abs_cmd)

        for _ in range(self.n_waits):
            # time.sleep(time_to_move)
            response = self.serial.read(1)
            # print(f"move response: {response}")
            if response == b"":
                time.sleep(self.wait_time)
            elif response == bytes.fromhex("0d"):
                self.safe_to_write.set()
                return True
        self.safe_to_write.set()
        return False

    # def set_relative_mode(self):
    #     """Set MP285 to Relative Position Mode.
    #
    #     This command sets the nature of the positional values specified with the Move
    #     (‘m’) command as relative positions as measured from the current position
    #     (absolute position returned by the Get Current Position (‘c’) command).
    #     The command sequence consists of 2 bytes: Command byte, followed by the
    #     terminator. Return data consists of 1 byte (task-complete indicator).
    #
    #     Returns
    #     -------
    #     command_complete : bool
    #         True if command was successful, False if not.
    #     """
    #     # print("calling set_relative_mode")
    #     self.flush_buffers()
    #     self.safe_write(bytes.fromhex("62") + bytes.fromhex("0d"))
    #     response = self.serial.read(1)
    #     if response == bytes.fromhex("0d"):
    #         command_complete = True
    #     else:
    #         command_complete = False
    #     self.safe_to_write.set()
    #     return command_complete

    def refresh_display(self):
        """Refresh the display on the MP-285 controller.

        This command refreshes the VFD (Vacuum Fluorescent Display) of the controller.
        The command sequence consists of 2 bytes: Command byte and terminator.
        Return data consists of 1 byte: task-complete indicator.

        Returns
        -------
        command_complete : bool
            True if command was successful, False if not.
        """
        # print("calling refresh_display")
        # self.flush_buffers()
        self.safe_write(bytes.fromhex("6E") + bytes.fromhex("0d"))
        response = self.serial.read(1)
        if response == bytes.fromhex("0d"):
            command_complete = True
        else:
            command_complete = False
        self.safe_to_write.set()
        return command_complete

    def reset_controller(self):
        """Reset the MP-285 controller.

        This command resets the controller. The command sequence consists of 2 bytes:
        Command byte and terminator. Return data consists of 1 byte: task-complete
        indicator.

        Returns
        -------
        command_complete : bool
            True if command was successful, False if not.
        """
        # print("calling reset_controller")
        # self.flush_buffers()
        self.safe_write(bytes.fromhex("72") + bytes.fromhex("0d"))
        response = self.serial.read(1)
        if response == bytes.fromhex("0d"):
            command_complete = True
        else:
            command_complete = False
        self.safe_to_write.set()
        return command_complete

    def get_controller_status(self):
        """Get the status of the MP-285 controller.

        This command gets status information from the controller and returns it in
        fixed-sized block of data. The command sequence consists of 2 bytes: Command
        byte and terminator. Return data consists of 33 bytes: 32 bytes of
        information and task-complete indicator.

        Returns
        -------
        command_complete = bool
            True if command was successful, False if not.
        """
        # print("calling get_controller_status")
        # self.flush_buffers()
        self.safe_write(bytes.fromhex("73") + bytes.fromhex("0d"))
        response = self.serial.read(33)
        if response[-1] == bytes.fromhex("0d"):
            command_complete = True
        else:
            command_complete = False

        # print(response)
        self.safe_to_write.set()
        # not implemented yet. See page 74 of documentation.
        return command_complete

    def close(self):
        """Close the serial connection to the stage"""
        self.serial.close()



# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only
# (subject to the limitations in the disclaimer below)
# provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Standard library imports
from tkinter import filedialog
import logging

# Third-party imports
# from matplotlib.figure import Figure
# from matplotlib.axes import Axes

# Local application imports
from navigate.controller.sub_controllers.gui_controller import GUIController

# Logger Setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)


class AdaptiveOpticsPopupController(GUIController):
    """Controller for the Adaptive Optics popup window"""

    def __init__(self, view, parent_controller):
        """Initialize the Adaptive Optics popup controller

        Parameters
        ----------
        view : AdaptiveOpticsPopupView
            The view for this controller
        parent_controller : Controller
            The parent controller for this controller
        """
        super().__init__(view, parent_controller)

        self.view.popup.protocol("WM_DELETE_WINDOW", self.view.popup.dismiss)

        self.parent_controller.configuration["experiment"]["AdaptiveOpticsParameters"][
            "HighlightedMode"
        ] = None

        #: list of dicts: list of dicts containing trace data
        self.trace_list = []

        #: dict: dictionary of widgets
        self.widgets = self.view.get_widgets()

        #: dict: dictionary of armed modes
        self.modes_armed = self.view.get_modes_armed()

        #: dict: dictionary of mode labels
        self.mode_labels = self.view.get_labels()

        # TODO: just for testing, remove later...
        #: list: list of cameras
        self.camera_list = self.view.camera_list
        self.camera_list[
            "values"
        ] = self.parent_controller.configuration_controller.microscope_list
        self.camera_list.bind(
            "<<ComboboxSelected>>", lambda evt: self.change_camera(evt)
        )

        #: Figure: figure for mirror plot
        self.fig = self.view.fig

        #: Figure: figure for tony wilson plot
        self.fig_tw = self.view.fig_tw

        #: Figure: figure for peaks plot
        self.peaks_plot = self.view.peaks_plot

        #: Figure: figure for trace plot
        self.trace_plot = self.view.trace_plot

        #: Figure: figure for mirror image
        self.mirror_img = self.view.mirror_img

        #: Figure: figure for coefs bar
        self.coefs_bar = self.view.coefs_bar

        # Trace Events
        self.view.set_button.configure(command=self.set_mirror)
        self.view.flat_button.configure(command=self.flatten_mirror)
        self.view.zero_button.configure(command=self.zero_mirror)
        self.view.clear_button.configure(command=self.clear_all_coefs)
        self.view.save_wcs_button.configure(command=self.save_wcs_file)
        self.view.from_wcs_button.configure(command=self.set_from_wcs_file)
        self.view.select_all_modes.configure(command=self.select_all_modes)
        self.view.deselect_all_modes.configure(command=self.deselect_all_modes)
        self.view.tony_wilson_button.configure(command=self.run_tony_wilson)

        for k in self.modes_armed.keys():
            self.modes_armed[k]["button"].configure(
                command=self.update_experiment_values
            )
            self.mode_labels[k].bind(
                "<Enter>", lambda evt, mode=k: self.set_highlighted_mode(evt, mode)
            )
            self.mode_labels[k].bind(
                "<Leave>", lambda evt: evt.widget.config(background="SystemButtonFace")
            )

        self.populate_experiment_values()

    def change_camera(self, evt):
        """Change the camera to the selected camera

        Parameters
        ----------
        evt : Event
            The event that triggered this function
        """
        # cam_id = evt.widget.get().split("_")[-1]
        # self.parent_controller.execute("change_camera", int(cam_id))
        cam_name = evt.widget.get()
        self.parent_controller.execute("resolution", cam_name)

    def set_highlighted_mode(self, evt, mode):
        """Set the highlighted mode

        Parameters
        ----------
        evt : Event
            The event that triggered this function
        mode : str
            The mode to highlight
        """
        evt.widget.config(background="red")
        self.parent_controller.configuration["experiment"]["AdaptiveOpticsParameters"][
            "HighlightedMode"
        ] = mode
        self.plot_tw_trace()

    def select_all_modes(self):
        """Select all modes"""
        for k in self.modes_armed.keys():
            self.modes_armed[k]["variable"].set(True)

    def deselect_all_modes(self):
        """Deselect all modes"""
        for k in self.modes_armed.keys():
            self.modes_armed[k]["variable"].set(False)

    def populate_experiment_values(self):
        """Populate the experiment values"""
        self.camera_list.set(
            self.parent_controller.configuration["experiment"]["MicroscopeState"][
                "microscope_name"
            ]
        )
        self.widgets["save_report"]["variable"].set(
            self.parent_controller.configuration["experiment"][
                "AdaptiveOpticsParameters"
            ].get("save_report", False)
        )
        coefs_dict = self.parent_controller.configuration["experiment"][
            "MirrorParameters"
        ]["modes"]
        tw_param_dict = self.parent_controller.configuration["experiment"][
            "AdaptiveOpticsParameters"
        ]["TonyWilson"]
        for k in coefs_dict.keys():
            self.widgets[k].set(coefs_dict[k])
        for k in tw_param_dict.keys():
            if k == "modes_armed":
                for mode_name in tw_param_dict["modes_armed"].keys():
                    self.modes_armed[mode_name]["variable"].set(
                        tw_param_dict["modes_armed"][mode_name]
                    )
            else:
                if type(self.widgets[k]) == dict:
                    self.widgets[k]["variable"].set(tw_param_dict[k])
                    continue

                self.widgets[k].set(tw_param_dict[k])

    def clear_all_coefs(self):
        """Clear all coefficients"""
        coefs_dict = self.parent_controller.configuration["experiment"][
            "MirrorParameters"
        ]["modes"]
        for k in coefs_dict.keys():
            self.widgets[k].set(0.0)
        self.update_experiment_values()

    def update_experiment_values(self):
        """Update the experiment values"""
        modes_dict = {}
        coef_list = self.get_coef_from_widgets()
        keys = self.view.mode_names
        for i, coef in enumerate(coef_list):
            modes_dict[keys[i]] = coef
        self.parent_controller.configuration["experiment"]["MirrorParameters"][
            "modes"
        ] = modes_dict

        self.parent_controller.configuration["experiment"]["AdaptiveOpticsParameters"][
            "TonyWilson"
        ]["iterations"] = int(self.widgets["iterations"].get())
        self.parent_controller.configuration["experiment"]["AdaptiveOpticsParameters"][
            "TonyWilson"
        ]["steps"] = int(self.widgets["steps"].get())
        self.parent_controller.configuration["experiment"]["AdaptiveOpticsParameters"][
            "TonyWilson"
        ]["amplitude"] = float(self.widgets["amplitude"].get())
        self.parent_controller.configuration["experiment"]["AdaptiveOpticsParameters"][
            "TonyWilson"
        ]["from"] = self.widgets["from"]["variable"].get()
        self.parent_controller.configuration["experiment"]["AdaptiveOpticsParameters"][
            "TonyWilson"
        ]["metric"] = self.widgets["metric"]["variable"].get()
        self.parent_controller.configuration["experiment"]["AdaptiveOpticsParameters"][
            "TonyWilson"
        ]["fitfunc"] = self.widgets["fitfunc"]["variable"].get()

        self.parent_controller.configuration["experiment"]["AdaptiveOpticsParameters"][
            "save_report"
        ] = self.widgets["save_report"]["variable"].get()

        for k in self.modes_armed.keys():
            self.parent_controller.configuration["experiment"][
                "AdaptiveOpticsParameters"
            ]["TonyWilson"]["modes_armed"][k] = self.modes_armed[k]["variable"].get()

        # print(self.parent_controller.configuration['experiment']['MirrorParameters']['modes'])

    def get_coef_from_widgets(self):
        """Get the coefficients from the widgets

        Returns
        -------
        list
            The coefficients
        """
        coef = []
        for k in self.widgets.keys():
            if k in self.view.mode_names:
                coef.append(float(self.widgets[k].get()))
        return coef

    def set_widgets_from_coef(self, coef):
        """Set the widgets from the coefficients

        Parameters
        ----------
        coef : list
            The coefficients
        """
        if list(coef):
            self.view.set_widgets(coef)
            self.update_experiment_values()

    def flatten_mirror(self):
        """Flatten the mirror"""
        self.parent_controller.execute("flatten_mirror")

    def zero_mirror(self):
        """Zero the mirror"""
        self.parent_controller.execute("zero_mirror")

    def set_mirror(self):
        """Set the mirror"""
        self.update_experiment_values()
        self.parent_controller.execute("set_mirror")

    def save_wcs_file(self):
        """Save the wcs file"""
        wcs_path = filedialog.asksaveasfilename(
            defaultextension=".wcs",
            initialdir="E:\\WaveKitX64\\MirrorFiles",
            filetypes=[("Wavefront File", "*.wcs")],
        )

        self.parent_controller.execute("save_wcs_file", wcs_path)

    def set_from_wcs_file(self):
        """Set the mirror from the wcs file"""
        wcs_path = filedialog.askopenfilename(
            defaultextension=".wcs",
            initialdir="E:\\WaveKitX64\\MirrorFiles",
            filetypes=[("Wavefront File", "*.wcs")],
        )

        self.parent_controller.execute("set_mirror_from_wcs", wcs_path)

    def run_tony_wilson(self):
        """Run the tony wilson routine"""
        self.update_experiment_values()
        # set the saving file name before running
        if self.widgets["save_report"]["variable"].get():
            save_dir = self.parent_controller.configuration["experiment"]["Saving"][
            "save_directory"
            ]

            self.report_path = filedialog.asksaveasfilename(
                defaultextension=".json",
                initialdir=save_dir,
                filetypes=[("JSON", "*.json")],
            )
        self.parent_controller.execute("tony_wilson")

    def save_report_to_file(self, report):
        print("Saving report!")

        full_report = {}

        for i, page in enumerate(report):
            full_report[f"iter_{i}"] = page       

        import pandas as pd

        try:
            pd.DataFrame(full_report).to_json(self.report_path)
        except Exception:
            pass

    def showup(self, popup_window=None):
        """Show the popup window."""
        # if popup_window is not None:
        #     self.view = popup_window
        self.view.popup.deiconify()
        self.view.popup.attributes("-topmost", 1)

    def plot_mirror(self, data):
        """Plot the mirror data.

        Parameters
        ----------
        data : dict
            The data to plot
        """

        try:
            self.mirror_img.clear()
            self.mirror_img.imshow(
                data["mirror_img"], cmap="bwr", vmin=-0.25, vmax=0.25
            )
            self.mirror_img.set_title("Mirror Pistons")
        except Exception:
            pass

        try:
            coefs = data["coefs"]
            self.coefs_bar.clear()
            self.coefs_bar.bar(range(len(coefs)), coefs)
            self.coefs_bar.set_title("Current Coefs")
            self.coefs_bar.set_xlabel("coef")
            self.coefs_bar.set_ylabel("amplitude")
        except Exception:
            pass

        # To redraw the plot
        self.fig.tight_layout()
        self.fig.canvas.draw_idle()

    def plot_tonywilson(self, data):
        """Plot the tony wilson data.

        Displays a plot of [focus, entropy] with data from autofocus routine

        Parameters
        ----------
        data : dict
            The data to plot
        """

        try:
            # Plotting data
            self.peaks_plot.clear()
            self.peaks_plot.plot(data["peaks"])
            self.peaks_plot.set_title("Image Metric")
            self.peaks_plot.set_xlabel("iteration")
            self.peaks_plot.set_ylabel(data["metric"])
        except KeyError:
            pass

        try:
            self.trace_list = data["trace"]
        except Exception:
            pass

        self.plot_tw_trace()

    def plot_tw_trace(self):
        """Plot the tony wilson trace data."""
        mode = self.parent_controller.configuration["experiment"][
            "AdaptiveOpticsParameters"
        ]["HighlightedMode"]

        if mode is not None:
            try:
                x = self.trace_list[mode]["x"]
                y = self.trace_list[mode]["y"]
                x_fit = self.trace_list[mode]["x_fit"]
                y_fit = self.trace_list[mode]["y_fit"]

                self.trace_plot.clear()
                self.trace_plot.plot(x, y, "*--")
                self.trace_plot.plot(x_fit, y_fit, color="r")
                self.trace_plot.set_title(f"Mode Fit: {mode}")
                self.trace_plot.set_xlabel("coef")
                self.trace_plot.set_ylabel("metric")
            except Exception:
                pass

        # To redraw the plot
        self.fig_tw.tight_layout()
        self.fig_tw.canvas.draw_idle()



# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only
# (subject to the limitations in the disclaimer below)
# provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Standard library imports
import logging

# Third-party imports

# Local application imports
from navigate.controller.sub_controllers.gui_controller import GUIController
from navigate.tools.file_functions import save_yaml_file
from navigate.tools.common_functions import combine_funcs
from navigate.config.config import update_config_dict

# Logger Setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)


class WaveformPopupController(GUIController):
    """Controller for the waveform popup window.

    This controller is responsible for the waveform popup window. It is responsible for
    updating the waveform constants in the configuration file and saving the waveform
    constants to a file.
    """

    def __init__(self, view, parent_controller, waveform_constants_path):
        """Initialize the WaveformPopupController.

        Parameters
        ----------
        view : WaveformPopup
            The view for the controller.
        parent_controller : Controller
            The parent controller.
        waveform_constants_path : str
            The path to the waveform constants file.
        """
        super().__init__(view, parent_controller)

        # Microscope information
        #: dict: Waveform constants for the microscope.
        self.resolution_info = self.parent_controller.configuration[
            "waveform_constants"
        ]
        #: dict: Galvo constants for the microscope.
        self.galvo_setting = self.resolution_info["galvo_constants"]
        #: ConfigurationController: The configuration controller.
        self.configuration_controller = self.parent_controller.configuration_controller
        #: str: The path to the waveform constants file.
        self.waveform_constants_path = waveform_constants_path

        # Get mode and mag widgets
        #: dict: The widgets for the mode and magnification.
        self.widgets = self.view.get_widgets()
        #: dict: The variables for the mode and magnification.
        self.variables = self.view.get_variables()

        # Get configuration
        #: list: The lasers.
        self.lasers = self.configuration_controller.lasers_info

        # Initialize
        #: str: The current resolution.
        self.resolution = None
        #: str: The current magnification.
        self.mag = None
        #: str: The current microscope operation mode.
        self.mode = "stop"
        #: dict: Remote focus experiment dictionary.
        self.remote_focus_experiment_dict = None
        #: bool: Flag to update galvo device.
        self.update_galvo_device_flag = None
        #: bool: Flag to update waveform parameters.
        self.update_waveform_parameters_flag = False
        #: bool: Flag to enable/disable waveforms.
        self.waveforms_enabled = True
        #: dict: Dictionary of amplitude values.
        self.amplitude_dict = None
        #: float: the minimum value of remote focus device
        self.laser_min = 0
        #: float: the maximum value of remote focus device
        self.laser_max = 1.0
        #: dict: Dictionary of galvo minimum values
        self.galvo_min = {}
        #: dict: Dictionary of galvo maximum values
        self.galvo_max = {}

        # event id list
        #: int: The event id.
        self.event_id = None

        # Event Binding
        # Switching microscopes modes (e.g., meso, nano, etc.)
        self.widgets["Mode"].widget.bind(
            "<<ComboboxSelected>>", self.show_magnification
        )

        # Switching magnifications (e.g., 10x, 20x, etc.)
        self.widgets["Mag"].widget.bind("<<ComboboxSelected>>", self.show_laser_info)

        # Changes to the waveform constants (amplitude, offset, etc.)
        for laser in self.lasers:
            self.variables[laser + " Amp"].trace_add(
                "write",
                self.update_remote_focus_settings(laser + " Amp", laser, "amplitude"),
            )
            self.variables[laser + " Off"].trace_add(
                "write",
                self.update_remote_focus_settings(laser + " Off", laser, "offset"),
            )

        # Changes to the galvo constants (amplitude, offset, etc.)
        for i in range(self.configuration_controller.galvo_num):
            galvo = f"Galvo {i}"
            self.variables[galvo + " Amp"].trace_add(
                "write", self.update_galvo_setting(galvo, " Amp", "amplitude")
            )
            self.variables[galvo + " Off"].trace_add(
                "write", self.update_galvo_setting(galvo, " Off", "offset")
            )
            self.variables[galvo + " Freq"].trace_add(
                "write", self.update_galvo_setting(galvo, " Freq", "frequency")
            )
            self.view.get_buttons()[galvo + " Freq"].configure(
                command=lambda: self.estimate_galvo_setting(galvo + " Freq")
            )

        # Changes in the delay, duty cycle, and smoothing waveform parameters
        # Delay, Duty, and Smoothing
        self.variables["Delay"].trace_add("write", self.update_waveform_parameters)
        self.variables["Duty"].trace_add("write", self.update_waveform_parameters)
        self.variables["Smoothing"].trace_add("write", self.update_waveform_parameters)

        # Save waveform constants
        self.view.get_buttons()["Save"].configure(command=self.save_waveform_constants)

        # Temporarily disable waveforms
        self.view.get_buttons()["toggle_waveform_button"].configure(
            command=self.toggle_waveform_state
        )

        # Save waveform constants upon closing the popup window
        self.view.popup.protocol(
            "WM_DELETE_WINDOW",
            combine_funcs(
                self.restore_amplitude,
                self.save_waveform_constants,
                self.view.popup.dismiss,
                lambda: delattr(self.parent_controller, "waveform_popup_controller"),
            ),
        )

        # Populate widgets
        self.widgets["Mode"].widget["values"] = list(
            self.resolution_info["remote_focus_constants"].keys()
        )
        self.widgets["Mode"].widget["state"] = "readonly"
        self.widgets["Mag"].widget["state"] = "readonly"

    def configure_widget_range(self):
        """Update widget ranges and precisions based on the current resolution mode.

        TODO: Other parameters we wish to enable/disable based on configuration?

        TODO: Should we instead change galvo amp/offset behavior based on a waveform
        type passed in the configuration? That is, should we pass galvo_l_waveform:
        sawtooth and galvo_r_waveform: dc_value? And then adjust the
        ETL_Popup_Controller accordingly? We could do the same for ETL vs. voice coil.

        This function updates the widget ranges and precisions based on the current
        resolution mode. The precision is set to -3 for high and nanoscale modes and -2
        for low mode. The increment is set to 0.001 for high and nanoscale modes and
        0.01 for low mode.

        """
        self.laser_min = self.configuration_controller.remote_focus_dict["hardware"]["min"]
        self.laser_max = self.configuration_controller.remote_focus_dict["hardware"]["max"]
        
        precision = int(self.configuration_controller.remote_focus_dict["hardware"].get("precision", 0))
        increment = int(self.configuration_controller.remote_focus_dict["hardware"].get("step", 0))
        if precision == 0:
            precision = -4 if self.laser_max < 1 else -3
        elif precision > 0:
            precision = - precision
        if increment == 0:
            increment = 0.0001 if self.laser_max < 1 else 0.001
        elif increment < 0:
            increment = - increment

        # set ranges of value for those lasers
        for laser in self.lasers:
            self.widgets[laser + " Amp"].widget.configure(from_=self.laser_min)
            self.widgets[laser + " Amp"].widget.configure(to=self.laser_max)
            self.widgets[laser + " Amp"].widget.configure(increment=increment)
            self.widgets[laser + " Amp"].widget.set_precision(precision)
            self.widgets[laser + " Amp"].widget.trigger_focusout_validation()
            # TODO: The offset bounds should adjust based on the amplitude bounds,
            #       so that amp + offset does not exceed the bounds. Can be done
            #       in update_remote_focus_settings()
            self.widgets[laser + " Off"].widget.configure(from_=self.laser_min)
            self.widgets[laser + " Off"].widget.configure(to=self.laser_max)
            self.widgets[laser + " Off"].widget.configure(increment=increment)
            self.widgets[laser + " Off"].widget.set_precision(precision)
            self.widgets[laser + " Off"].widget.trigger_focusout_validation()

        for galvo, d in zip(self.galvos, self.galvo_dict):
            galvo_min = d["hardware"]["min"]
            galvo_max = d["hardware"]["max"]
            self.widgets[galvo + " Amp"].widget.configure(from_=galvo_min)
            self.widgets[galvo + " Amp"].widget.configure(to=galvo_max)
            self.widgets[galvo + " Amp"].widget.configure(increment=increment)
            self.widgets[galvo + " Amp"].widget.set_precision(precision)
            self.widgets[galvo + " Amp"].widget["state"] = "normal"
            self.widgets[galvo + " Amp"].widget.trigger_focusout_validation()
            # TODO: The offset bounds should adjust based on the amplitude bounds,
            #       so that amp + offset does not exceed the bounds. Can be done
            #       in update_remote_focus_settings()
            self.widgets[galvo + " Off"].widget.configure(from_=galvo_min)
            self.widgets[galvo + " Off"].widget.configure(to=galvo_max)
            self.widgets[galvo + " Off"].widget.configure(increment=increment)
            self.widgets[galvo + " Off"].widget.set_precision(precision)
            self.widgets[galvo + " Off"].widget["state"] = "normal"
            self.widgets[galvo + " Off"].widget.trigger_focusout_validation()

            self.widgets[galvo + " Freq"].widget.configure(from_=0)
            self.widgets[galvo + " Freq"].widget.configure(increment=increment)
            self.widgets[galvo + " Freq"].widget.set_precision(precision)
            self.widgets[galvo + " Freq"].widget["state"] = "normal"
            self.widgets[galvo + " Freq"].widget.trigger_focusout_validation()

            self.galvo_min[galvo] = galvo_min
            self.galvo_max[galvo] = galvo_max

        for i in range(len(self.galvos), self.configuration_controller.galvo_num):
            galvo_name = f"Galvo {i}"
            self.widgets[galvo_name + " Amp"].widget["state"] = "disabled"
            self.widgets[galvo_name + " Off"].widget["state"] = "disabled"
            self.widgets[galvo_name + " Freq"].widget["state"] = "disabled"

        # The galvo by default uses a sawtooth waveform.
        # However, sometimes we have a resonant galvo.
        # In the case of the resonant galvo, amplitude is zero and only the offset
        # can be controlled. We only define the offset in the configuration.yml file.
        # If only the offset is defined for galvo_{focus_prefix}, we disable the
        # amplitude.
        #

    def populate_experiment_values(self):
        """Set experiment values."""
        self.remote_focus_experiment_dict = self.parent_controller.configuration[
            "experiment"
        ]["MicroscopeState"]
        resolution_value = self.remote_focus_experiment_dict["microscope_name"]
        zoom_value = self.remote_focus_experiment_dict["zoom"]
        mag = zoom_value
        if (
            self.widgets["Mode"].get() == resolution_value
            and self.widgets["Mag"].get() == mag
        ):
            return
        self.widgets["Mode"].set(resolution_value)
        self.show_magnification(mag)

    def showup(self):
        """This function will let the popup window show in front."""
        self.view.popup.deiconify()
        self.view.popup.attributes("-topmost", 1)

    def show_magnification(self, *args):
        """Show magnification options when the user changes the focus mode.

        Parameters
        ----------
        *args : tuple
            The first element is the new focus mode.
        """
        # restore amplitude before change resolution if needed
        self.restore_amplitude()
        # get resolution setting
        self.resolution = self.widgets["Mode"].widget.get()
        temp = list(
            self.resolution_info["remote_focus_constants"][self.resolution].keys()
        )
        self.widgets["Mag"].widget["values"] = temp

        if args[0] in temp:
            self.widgets["Mag"].widget.set(args[0])
        else:
            self.widgets["Mag"].widget.set(temp[0])

        # update laser info
        self.show_laser_info()

    def show_laser_info(self, *args):
        """Show laser info when the user changes magnification setting.

        Parameters
        ----------
        *args : tuple
            The first element is the new magnification setting.
        """
        # get galvo dict for the specified microscope/magnification
        self.galvo_dict = self.parent_controller.configuration["configuration"][
            "microscopes"
        ][self.resolution]["galvo"]
        self.galvos = [f"Galvo {i}" for i in range(len(self.galvo_dict))]
        # restore amplitude before change mag if needed
        self.restore_amplitude()
        # get magnification setting
        self.mag = self.widgets["Mag"].widget.get()
        for laser in self.lasers:
            self.variables[laser + " Amp"].set(
                self.resolution_info["remote_focus_constants"][self.resolution][
                    self.mag
                ][laser]["amplitude"]
            )
            self.variables[laser + " Off"].set(
                self.resolution_info["remote_focus_constants"][self.resolution][
                    self.mag
                ][laser]["offset"]
            )

        # do not tell the model to update galvo
        self.update_galvo_device_flag = False
        for galvo in self.galvos:
            self.variables[galvo + " Amp"].set(
                self.galvo_setting[galvo][self.resolution][self.mag].get("amplitude", 0)
            )
            self.variables[galvo + " Off"].set(
                self.galvo_setting[galvo][self.resolution][self.mag].get("offset", 0)
            )
            self.variables[galvo + " Freq"].set(
                self.galvo_setting[galvo][self.resolution][self.mag].get("frequency", 0)
            )
        self.update_galvo_device_flag = True

        # Load waveform parameters from configuration - Smooth, Delay, Duty Cycle.
        # Provide defaults should loading fail.
        # Currently pulls from the original microscope configuration YAML file, not the
        # current microscope configuration according to the model
        self.update_waveform_parameters_flag = False
        waveform_configuration = self.resolution_info["remote_focus_constants"][
            self.resolution
        ][self.mag][self.lasers[0]]
        self.widgets["Smoothing"].set(
            waveform_configuration.get("percent_smoothing", 0)
        )
        self.widgets["Smoothing"].widget.trigger_focusout_validation()
        self.widgets["Delay"].set(waveform_configuration.get("percent_delay", 7.5))
        self.widgets["Delay"].widget.trigger_focusout_validation()
        if "other_constants" not in self.resolution_info:
            update_config_dict(
                self.parent_controller.manager,
                self.resolution_info,
                "other_constants",
                {"remote_focus_settle_duration": 0},
            )
        self.widgets["Duty"].set(
            self.resolution_info["other_constants"]["remote_focus_settle_duration"]
        )
        self.widgets["Duty"].widget.trigger_focusout_validation()
        self.update_waveform_parameters_flag = True

        # update resolution value in central controller (menu)
        value = f"{self.resolution} {self.mag}"
        if self.parent_controller.menu_controller.resolution_value.get() != value:
            self.parent_controller.menu_controller.resolution_value.set(value)

        # reconfigure widgets
        self.configure_widget_range()

    def update_remote_focus_settings(self, name, laser, remote_focus_name):
        """Update remote focus settings in memory.

        Parameters
        ----------
        name : str
            The name of the variable.
        laser : str
            The name of the laser.
        remote_focus_name : str
            The name of the remote focus setting.
        """
        variable = self.variables[name]

        # TODO: Is this still a bug?
        # BUG Upon startup this will always run 0.63x,
        # and when changing magnification it will run 0.63x
        # before whatever mag is selected
        def func_laser(*args):
            value = self.resolution_info["remote_focus_constants"][self.resolution][
                self.mag
            ][laser][remote_focus_name]

            # Will only run code if value in constants does not match whats in GUI
            # for Amp or Off AND in Live mode
            # TODO: Make also work in the 'single' acquisition mode.
            variable_value = variable.get()
            logger.debug(
                f"Remote Focus Amplitude/Offset Changed pre if statement: "
                f"{variable_value}"
            )
            if value != variable_value and variable_value != "":
                # tell parent controller (the device)
                if self.event_id:
                    self.view.popup.after_cancel(self.event_id)
                
                try:
                    value = float(variable_value)
                except ValueError:
                    return
                if value < self.laser_min or value > self.laser_max:
                    return
                self.resolution_info["remote_focus_constants"][self.resolution][
                    self.mag
                ][laser][remote_focus_name] = variable_value
                logger.debug(
                    f"Remote Focus Amplitude/Offset Changed:, {variable_value}"
                )

                # Delay feature.
                self.event_id = self.view.popup.after(
                    500,
                    lambda: self.parent_controller.execute(
                        "update_setting", "resolution"
                    ),
                )

        return func_laser

    def update_waveform_parameters(self, *args, **wargs):
        """Update the waveform parameters for delay, duty cycle, and smoothing.

        Communicate changes to the parent controller.

        Parameters
        ----------
        *args : tuple
            The first element is the new waveform.
        **wargs : dict
            The key is the name of the waveform and the value is the waveform
        """
        if not self.update_waveform_parameters_flag:
            return
        
        if self.event_id:
            self.view.popup.after_cancel(self.event_id)
        # Get the values from the widgets.
        try:
            delay = float(self.widgets["Delay"].widget.get())
            duty_cycle = float(self.widgets["Duty"].widget.get())
            smoothing = float(self.widgets["Smoothing"].widget.get())
        except ValueError:
            return

        # Update the waveform parameters
        # all the lasers use the same delay, smoothing value
        for laser in self.lasers:
            self.resolution_info["remote_focus_constants"][self.resolution][self.mag][
                laser
            ]["percent_delay"] = delay
            self.resolution_info["remote_focus_constants"][self.resolution][self.mag][
                laser
            ]["percent_smoothing"] = smoothing
        self.resolution_info["other_constants"][
            "remote_focus_settle_duration"
        ] = duty_cycle

        # Pass the values to the parent controller.
        self.event_id = self.view.popup.after(
            500,
            lambda: self.parent_controller.execute(
                "update_setting", "waveform_parameters"
            ),
        )

    def estimate_galvo_setting(self, *args, **kwargs):
        """Estimate galvo settings according to the acquisition parameters.

        Will only work if all channels have the same exposure duration.
        Gets the line interval from the camera, number of pixels from the light-sheet
        mode, and estimates the frequency as 1 / (line interval * number of pixels).

        Parameters
        ----------
        *args : tuple
            The first element is the name of the galvo.
        **kwargs : dict
            The key is the name of the galvo and the value is the galvo setting.
        """

        galvo_name = args[0]

        number_of_pixels = (
            self.parent_controller.camera_setting_controller.mode_widgets[
                "Pixels"
            ].get()
        )
        if number_of_pixels == "":
            # If we are not in the light-sheet mode, widget returns an empty string.
            return

        # TODO: Worth noting that this doesn't matter because it cancels out in the
        # frequency bit, but this is often not the right value. The framerate widget
        # doesn't update unless you change the exposure time in the channel settings.
        # It launches as a default of 100 ms every time.
        exposure_time = (
            self.parent_controller.camera_setting_controller.framerate_widgets[
                "exposure_time"
            ].get()
        )

        # The camera line interval won't be set until starting acquisition,
        # we need to calculate it directly
        # exposure_time and light_sheet_exposure_time are both ms
        (
            light_sheet_exposure_time,
            _,
        ) = self.parent_controller.model.get_camera_line_interval_and_exposure_time(
            exposure_time, int(number_of_pixels) + 1
        )  # TODO: Unclear why we need the +1. Figure out the reason.

        frequency = 2 / light_sheet_exposure_time * exposure_time

        # Update the GUI
        self.view.inputs[galvo_name].widget.set(round(frequency, 3))

    def update_galvo_setting(self, galvo_name, widget_name, parameter):
        """Update galvo settings in memory.

        Parameters
        ----------
        galvo_name : str
            The name of the galvo.
        widget_name : str
            The name of the widget.
        parameter : str
            The name of the parameter.

        Returns
        -------
        func_galvo : function
            The function to update the galvo setting.
        """
        name = galvo_name + widget_name
        variable = self.variables[name]

        def func_galvo(*args):
            if not self.update_galvo_device_flag:
                return
            try:
                value = self.galvo_setting[galvo_name][self.resolution][self.mag][
                    parameter
                ]
            except KeyError:
                # Special case for galvo amplitude not being defined
                value = 0
            variable_value = variable.get()
            logger.debug(
                f"Galvo parameter {parameter} changed: "
                f"{variable_value} pre if statement"
            )
            if value != variable_value and variable_value != "":
                # change any galvo parameters as one event
                if self.event_id:
                    self.view.popup.after_cancel(self.event_id)
                
                try:
                    value = float(variable_value)
                except ValueError:
                    return
                if value < self.galvo_min[galvo_name] or value > self.galvo_max[galvo_name]:
                    return
                self.galvo_setting[galvo_name][self.resolution][self.mag][
                    parameter
                ] = variable_value
                logger.debug(f"Galvo parameter {parameter} changed: {variable_value}")
                
                self.event_id = self.view.popup.after(
                    500,
                    lambda: self.parent_controller.execute("update_setting", "galvo"),
                )

        return func_galvo

    def save_waveform_constants(self):
        """Save updated waveform parameters to yaml file."""
        # errors = self.get_errors()
        # if errors:
        #     return  # Dont save if any errors TODO needs testing
        save_yaml_file("", self.resolution_info, self.waveform_constants_path)

    """
    Example for preventing submission of a field/controller. So if there is an error in
    any field that is supposed to have validation then the config cannot be saved.
    """
    # TODO needs testing may also need to be moved to the remote_focus_popup class.
    #  Opinions welcome
    # def get_errors(self):
    #     """
    #     Get a list of field errors in popup
    #     """

    #     errors = {}
    #     for key, labelInput in self.widgets.items():
    #         if hasattr(labelInput.widget, 'trigger_focusout_validation'):
    #             labelInput.widget.trigger_focusout_validation()
    #         if labelInput.error.get():
    #             errors[key] = labelInput.error.get()
    #     return errors
    def toggle_waveform_state(self):
        """Temporarily disable waveform amplitude for quick alignment on stationary
        beam.
        """
        if self.waveforms_enabled is True:
            self.view.buttons["toggle_waveform_button"].config(state="disabled")
            self.view.buttons["toggle_waveform_button"].config(text="Enable Waveforms")
            self.amplitude_dict = {"resolution": self.resolution, "mag": self.mag}

            for laser in self.lasers:
                self.amplitude_dict[laser] = self.resolution_info[
                    "remote_focus_constants"
                ][self.resolution][self.mag][laser]["amplitude"]
                self.variables[laser + " Amp"].set(0)
                self.widgets[laser + " Amp"].widget.config(state="disabled")

            for galvo in self.galvos:
                self.amplitude_dict[galvo] = self.resolution_info["galvo_constants"][
                    galvo
                ][self.resolution][self.mag]["amplitude"]
                self.variables[galvo + " Amp"].set(0)
                self.widgets[galvo + " Amp"].widget.config(state="disabled")

            # Need to update main controller.
            self.waveforms_enabled = False
            self.view.popup.after(
                500,
                lambda: self.view.buttons["toggle_waveform_button"].config(
                    state="normal"
                ),
            )
        else:
            self.view.buttons["toggle_waveform_button"].config(state="disabled")
            self.show_laser_info()
            # call the parent controller the amplitude values are updated
            try:
                if self.event_id:
                    self.view.popup.after_cancel(self.event_id)
            except KeyError:
                pass

            self.event_id = self.view.popup.after(
                500,
                lambda: self.parent_controller.execute("update_setting", "galvo"),
            )
            self.view.popup.after(
                500,
                lambda: self.view.buttons["toggle_waveform_button"].config(
                    state="normal"
                ),
            )

    def restore_amplitude(self):
        """Restore amplitude values to previous values."""
        self.view.buttons["toggle_waveform_button"].config(text="Disable Waveforms")
        self.waveforms_enabled = True
        if self.amplitude_dict is None:
            return
        resolution = self.amplitude_dict["resolution"]
        mag = self.amplitude_dict["mag"]
        for laser in self.lasers:
            self.resolution_info["remote_focus_constants"][resolution][mag][laser][
                "amplitude"
            ] = self.amplitude_dict[laser]
            self.widgets[laser + " Amp"].widget.config(state="normal")
        for galvo in self.galvos:
            self.resolution_info["galvo_constants"][galvo][resolution][mag][
                "amplitude"
            ] = self.amplitude_dict[galvo]
            self.widgets[galvo + " Amp"].widget.config(state="normal")
        self.amplitude_dict = None
